# Copyright 2022 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for testutils._hmc_definitions module.
"""


import os
import tempfile
import pytest

from zhmcclient.testutils._hmc_definitions import HMCDefinitions, \
    HMCNoVaultError, HMCNotFound

# Test cases for all methods and properties of HMCDefinitions.
TESTCASES_HMC_DEFINITIONS_ALL = [
    # Format:
    # * desc: Testcase description
    # * inv_data: String with data of the HMC inventory file to test
    # * vault_data: String with data of the HMC vault file to test
    # * testhmc: Group or host name of the HMCs to be tested
    # * exp_hmc_defs: Dict of expected HMCDefinition properties
    # * exp_init_exc_type: Expected exception type during init, or None
    # * exp_list_exc_type: Expected exception type during list, or None
    (
        "Empty inventory and vault files",
        """
all:
  hosts:
        """,
        """
hmc_auth:
        """,
        None,
        {
            "all": {},
        },
        None, None
    ),
    (
        "One real HMC with all properties",
        """
all:
  hosts:
    host1:
      description: host1
      contact: contact1
      access_via: Intranet
      ansible_host: 1.2.3.1
      cpcs:
        CPC1:
          dpm_enabled: true
        """,
        """
hmc_auth:
  host1:
    userid: userid1
    password: password1
    verify: true
    ca_certs: mycerts.pem
        """,
        None,
        {
            "all": {
                "host1": {
                    "description": "host1",
                    "contact": "contact1",
                    "access_via": "Intranet",
                    "mock_file": None,
                    "host": "1.2.3.1",
                    "userid": "userid1",
                    "password": "password1",
                    "verify": True,
                    "ca_certs": "mycerts.pem",
                    "cpcs": {
                        "CPC1": {
                            "dpm_enabled": True,
                        },
                    },
                    "add_vars": {},
                },
            },
        },
        None, None
    ),
    (
        "One real HMC with minimal properties",
        """
all:
  hosts:
    host1:
      ansible_host: 1.2.3.1
        """,
        """
hmc_auth:
  host1:
    userid: userid1
    password: password1
        """,
        None,
        {
            "all": {
                "host1": {
                    "description": "",
                    "contact": "",
                    "access_via": "",
                    "host": "1.2.3.1",
                    "mock_file": None,
                    "userid": "userid1",
                    "password": "password1",
                    "verify": True,
                    "ca_certs": None,
                    "cpcs": {},
                    "add_vars": {},
                },
            },
        },
        None, None
    ),
    (
        "One mocked HMC with all properties",
        """
all:
  hosts:
    host1:
      description: host1
      contact: contact1
      access_via: Intranet
      mock_file: host1_mock.yaml
      cpcs:
        CPC1:
          dpm_enabled: true
        """,
        """
hmc_auth:
  host1:
    userid: userid1
    password: password1
        """,
        None,
        {
            "all": {
                "host1": {
                    "description": "host1",
                    "contact": "contact1",
                    "access_via": "Intranet",
                    "host": None,
                    "mock_file": "host1_mock.yaml",
                    "userid": "userid1",
                    "password": "password1",
                    "verify": True,
                    "ca_certs": None,
                    "cpcs": {
                        "CPC1": {
                            "dpm_enabled": True,
                        },
                    },
                    "add_vars": {},
                },
            },
        },
        None, None
    ),
    (
        "One mocked HMC with minimal properties",
        """
all:
  hosts:
    host1:
      mock_file: host1_mock.yaml
        """,
        """
hmc_auth:
  host1:
    userid: userid1
    password: password1
        """,
        None,
        {
            "all": {
                "host1": {
                    "description": "",
                    "contact": "",
                    "access_via": "",
                    "host": None,
                    "mock_file": "host1_mock.yaml",
                    "userid": "userid1",
                    "password": "password1",
                    "verify": True,
                    "ca_certs": None,
                    "cpcs": {},
                    "add_vars": {},
                },
            },
        },
        None, None
    ),
    (
        "One real HMC that is used in a child group without overridden props",
        """
all:
  hosts:
    host1:
      description: desc1
      ansible_host: 1.2.3.1
  children:
    group1:
      hosts:
        host1:
        """,
        """
hmc_auth:
  host1:
    userid: userid1
    password: password1
        """,
        None,
        {
            "all": {
                "host1": {
                    "description": "desc1",
                    "host": "1.2.3.1",
                    "userid": "userid1",
                    "password": "password1",
                },
            },
            "group1": {
                "host1": {
                    "description": "desc1",
                    "host": "1.2.3.1",
                    "userid": "userid1",
                    "password": "password1",
                },
            },
        },
        None, None
    ),
    (
        "One real HMC that is used in a child group and overrides all props",
        """
all:
  hosts:
    host1:
      description: desc_parent
      ansible_host: 1.2.3.1
      contact: contact_parent
      access_via: access_via_parent
      cpcs:
        CPC1:
          dpm_enabled: true
  children:
    group1:
      hosts:
        host1:
          description: desc_child
          ansible_host: 2.2.3.1
        """,
        """
hmc_auth:
  host1:
    userid: userid1
    password: password1
    verify: true
    ca_certs: mycerts.pem
        """,
        None,
        {
            "all": {
                "host1": {
                    "description": "desc_parent",
                    "host": "1.2.3.1",
                    "contact": "contact_parent",
                    "access_via": "access_via_parent",
                    "cpcs": {
                        "CPC1": {
                            "dpm_enabled": True,
                        },
                    },
                    "add_vars": {},
                    "userid": "userid1",
                    "password": "password1",
                    "verify": True,
                    "ca_certs": "mycerts.pem",
                },
            },
            "group1": {
                "host1": {
                    "description": "desc_child",
                    "host": "2.2.3.1",
                    "contact": "contact_parent",
                    "access_via": "access_via_parent",
                    "cpcs": {
                        "CPC1": {
                            "dpm_enabled": True,
                        },
                    },
                    "add_vars": {},
                    "userid": "userid1",
                    "password": "password1",
                    "verify": True,
                    "ca_certs": "mycerts.pem",
                },
            },
        },
        None, None
    ),
    (
        "One mocked HMC that is used in a child group without overridden props",
        """
all:
  hosts:
    host1:
      description: desc1
      mock_file: mymock.yaml
  children:
    group1:
      hosts:
        host1:
        """,
        """
hmc_auth:
  host1:
    userid: userid1
    password: password1
        """,
        None,
        {
            "all": {
                "host1": {
                    "description": "desc1",
                    "host": None,
                    "mock_file": "mymock.yaml",
                    "userid": "userid1",
                    "password": "password1",
                },
            },
            "group1": {
                "host1": {
                    "description": "desc1",
                    "host": None,
                    "mock_file": "mymock.yaml",
                    "userid": "userid1",
                    "password": "password1",
                },
            },
        },
        None, None
    ),
    (
        "One real HMC with missing corresponding vault file entry",
        """
all:
  hosts:
    host1:
      ansible_host: 1.2.3.1
        """,
        """
hmc_auth:
  host2:
    userid: userid2
    password: password2
        """,
        None,
        None,
        HMCNoVaultError, None
    ),
    (
        "One real HMC with incorrect testhmc",
        """
all:
  hosts:
    host1:
      ansible_host: 1.2.3.1
        """,
        """
hmc_auth:
  host1:
    userid: userid1
    password: password1
        """,
        "host2",
        None,
        None, HMCNotFound
    ),
    (
        "One host and one child group with a second host",
        """
all:
  hosts:
    host1:
      description: host1
      ansible_host: 1.2.3.1
  children:
    east:
      hosts:
        1.2.3.2:
          description: host2
        """,
        """
hmc_auth:
  host1:
    userid: userid1
    password: password1
  1.2.3.2:
    userid: userid2
    password: password2
        """,
        'host1',
        {
            "all": {
                "host1": {
                    "description": "host1",
                    "contact": "",
                    "access_via": "",
                    "host": "1.2.3.1",
                    "mock_file": None,
                    "userid": "userid1",
                    "password": "password1",
                    "verify": True,
                    "ca_certs": None,
                    "cpcs": {},
                    "add_vars": {},
                },
                "1.2.3.2": {
                    "description": "host2",
                    "contact": "",
                    "access_via": "",
                    "host": "1.2.3.2",
                    "mock_file": None,
                    "userid": "userid2",
                    "password": "password2",
                    "verify": True,
                    "ca_certs": None,
                    "cpcs": {},
                    "add_vars": {},
                },
            },
            "east": {
                "1.2.3.2": {
                    "description": "host2",
                    "contact": "",
                    "access_via": "",
                    "host": "1.2.3.2",
                    "mock_file": None,
                    "userid": "userid2",
                    "password": "password2",
                    "verify": True,
                    "ca_certs": None,
                    "cpcs": {},
                    "add_vars": {},
                },
            },
        },
        None, None
    ),
    (
        "Additional variables at top level host",
        """
all:
  hosts:
    host1:
      description: host1
      ansible_host: 1.2.3.1
      var1: value1
        """,
        """
hmc_auth:
  host1:
    userid: userid1
    password: password1
        """,
        'host1',
        {
            "all": {
                "host1": {
                    "description": "host1",
                    "contact": "",
                    "access_via": "",
                    "host": "1.2.3.1",
                    "mock_file": None,
                    "userid": "userid1",
                    "password": "password1",
                    "verify": True,
                    "ca_certs": None,
                    "cpcs": {},
                    "add_vars": {
                        "var1": "value1",
                    },
                },
            },
        },
        None, None
    ),
    (
        "Variables at top level propagated to top-level hosts",
        """
all:
  hosts:
    host1:
      description: host1
      ansible_host: 1.2.3.1
  vars:
    var1: value1
        """,
        """
hmc_auth:
  host1:
    userid: userid1
    password: password1
        """,
        'host1',
        {
            "all": {
                "host1": {
                    "description": "host1",
                    "contact": "",
                    "access_via": "",
                    "host": "1.2.3.1",
                    "mock_file": None,
                    "userid": "userid1",
                    "password": "password1",
                    "verify": True,
                    "ca_certs": None,
                    "cpcs": {},
                    "add_vars": {
                        "var1": "value1",
                    },
                },
            },
        },
        None, None
    ),
    (
        "Variables at top level propagated to top-level hosts and child hosts",
        """
all:
  hosts:
    host1:
      description: host1
      ansible_host: 1.2.3.1
  vars:
    var1: value1
  children:
    east:
      hosts:
        host1:
        """,
        """
hmc_auth:
  host1:
    userid: userid1
    password: password1
        """,
        'host1',
        {
            "all": {
                "host1": {
                    "description": "host1",
                    "contact": "",
                    "access_via": "",
                    "host": "1.2.3.1",
                    "mock_file": None,
                    "userid": "userid1",
                    "password": "password1",
                    "verify": True,
                    "ca_certs": None,
                    "cpcs": {},
                    "add_vars": {
                        "var1": "value1",
                    },
                },
            },
            "east": {
                "host1": {
                    "description": "host1",
                    "contact": "",
                    "access_via": "",
                    "host": "1.2.3.1",
                    "mock_file": None,
                    "userid": "userid1",
                    "password": "password1",
                    "verify": True,
                    "ca_certs": None,
                    "cpcs": {},
                    "add_vars": {
                        "var1": "value1",
                    },
                },
            },
        },
        None, None
    ),
    (
        "Variables at top level overridden at child group level and propagated",
        """
all:
  hosts:
    host1:
      description: host1
      ansible_host: 1.2.3.1
  vars:
    var1: value1
  children:
    east:
      hosts:
        host1:
      vars:
        var1: value2
        """,
        """
hmc_auth:
  host1:
    userid: userid1
    password: password1
        """,
        'host1',
        {
            "all": {
                "host1": {
                    "description": "host1",
                    "contact": "",
                    "access_via": "",
                    "host": "1.2.3.1",
                    "mock_file": None,
                    "userid": "userid1",
                    "password": "password1",
                    "verify": True,
                    "ca_certs": None,
                    "cpcs": {},
                    "add_vars": {
                        "var1": "value1",
                    },
                },
            },
            "east": {
                "host1": {
                    "description": "host1",
                    "contact": "",
                    "access_via": "",
                    "host": "1.2.3.1",
                    "mock_file": None,
                    "userid": "userid1",
                    "password": "password1",
                    "verify": True,
                    "ca_certs": None,
                    "cpcs": {},
                    "add_vars": {
                        "var1": "value2",
                    },
                },
            },
        },
        None, None
    ),
    (
        "Variables at top level overridden at top host level",
        """
all:
  hosts:
    host1:
      description: host1
      ansible_host: 1.2.3.1
      var1: value2
  vars:
    var1: value1
        """,
        """
hmc_auth:
  host1:
    userid: userid1
    password: password1
        """,
        'host1',
        {
            "all": {
                "host1": {
                    "description": "host1",
                    "contact": "",
                    "access_via": "",
                    "host": "1.2.3.1",
                    "mock_file": None,
                    "userid": "userid1",
                    "password": "password1",
                    "verify": True,
                    "ca_certs": None,
                    "cpcs": {},
                    "add_vars": {
                        "var1": "value2",
                    },
                },
            },
        },
        None, None
    ),
    (
        "Variables at top level overridden at child group level and again at "
        "child host level",
        """
all:
  hosts:
    host1:
      description: host1
      ansible_host: 1.2.3.1
  vars:
    var1: value1
  children:
    east:
      hosts:
        host1:
          var1: value3
      vars:
        var1: value2
        """,
        """
hmc_auth:
  host1:
    userid: userid1
    password: password1
        """,
        'host1',
        {
            "all": {
                "host1": {
                    "description": "host1",
                    "contact": "",
                    "access_via": "",
                    "host": "1.2.3.1",
                    "mock_file": None,
                    "userid": "userid1",
                    "password": "password1",
                    "verify": True,
                    "ca_certs": None,
                    "cpcs": {},
                    "add_vars": {
                        "var1": "value1",
                    },
                },
            },
            "east": {
                "host1": {
                    "description": "host1",
                    "contact": "",
                    "access_via": "",
                    "host": "1.2.3.1",
                    "mock_file": None,
                    "userid": "userid1",
                    "password": "password1",
                    "verify": True,
                    "ca_certs": None,
                    "cpcs": {},
                    "add_vars": {
                        "var1": "value3",
                    },
                },
            },
        },
        None, None
    ),
]


@pytest.mark.parametrize(
    "desc, inv_data, vault_data, testhmc, exp_hmc_defs, exp_init_exc_type, "
    "exp_list_exc_type",
    TESTCASES_HMC_DEFINITIONS_ALL)
def test_HMCDefinitions_all(
        desc, inv_data, vault_data, testhmc, exp_hmc_defs, exp_init_exc_type,
        exp_list_exc_type):
    # pylint: disable=unused-argument,invalid-name
    """
    Test function for all methods and properties of HMCDefinitions.
    """

    with tempfile.NamedTemporaryFile(
            mode='w', suffix='.yaml', delete=False) as tf_inv:
        tf_inv.write(inv_data)

    with tempfile.NamedTemporaryFile(
            mode='w', suffix='.yaml', delete=False) as tf_vault:
        tf_vault.write(vault_data)

    try:
        if exp_init_exc_type:
            with pytest.raises(exp_init_exc_type) as exc_info:

                # The function to be tested
                HMCDefinitions(
                    inventory_file=tf_inv.name, vault_file=tf_vault.name,
                    testhmc=testhmc, load=True)

            exc = exc_info.value
            assert isinstance(exc, exp_init_exc_type)

        elif exp_list_exc_type:

            # The function to be tested
            hmc_defs = HMCDefinitions(
                inventory_file=tf_inv.name, vault_file=tf_vault.name,
                testhmc=testhmc, load=True)

            with pytest.raises(exp_list_exc_type) as exc_info:

                # The function to be tested
                hmc_defs.list_hmcs(testhmc)

            exc = exc_info.value
            assert isinstance(exc, exp_list_exc_type)

        else:  # success

            # The function to be tested
            hmc_defs = HMCDefinitions(
                inventory_file=tf_inv.name, vault_file=tf_vault.name,
                testhmc=testhmc, load=True)

            # The function to be tested
            group_names = hmc_defs.group_names

            for group_name in group_names:

                # The function to be tested
                hd_list = hmc_defs.list_hmcs(group_name)

                assert_hmcdefs(group_name, hd_list, exp_hmc_defs)

            # The function to be tested
            hd_list = hmc_defs.list_all_hmcs()

            assert_hmcdefs('all', hd_list, exp_hmc_defs)

    finally:
        os.remove(tf_inv.name)
        os.remove(tf_vault.name)


def assert_hmcdefs(group_name, hd_list, exp_hmc_defs):
    """
    Check a HMC definition list for a particular group name.

    This accesses and thus tests further properties of the HMCDefinitions class.
    """
    assert group_name in exp_hmc_defs

    exp_props_dict = exp_hmc_defs[group_name]
    for hd in hd_list:

        # The function to be tested
        nickname = hd.nickname
        assert nickname in exp_props_dict

        exp_props = exp_props_dict[nickname]

        if 'description' in exp_props:
            # The function to be tested
            description = hd.description
            assert description == exp_props['description']

        if 'contact' in exp_props:
            # The function to be tested
            contact = hd.contact
            assert contact == exp_props['contact']

        if 'host' in exp_props:
            # The function to be tested
            host = hd.host
            assert host == exp_props['host']

        if 'userid' in exp_props:
            # The function to be tested
            userid = hd.userid
            assert userid == exp_props['userid']

        if 'password' in exp_props:
            # The function to be tested
            password = hd.password
            assert password == exp_props['password']

        if 'verify' in exp_props:
            # The function to be tested
            verify = hd.verify
            assert verify == exp_props['verify']

        if 'ca_certs' in exp_props:
            # The function to be tested
            ca_certs = hd.ca_certs
            assert ca_certs == exp_props['ca_certs']

        if 'cpcs' in exp_props:
            # The function to be tested
            cpcs = hd.cpcs
            assert cpcs == exp_props['cpcs']

        if 'add_vars' in exp_props:
            # The function to be tested
            add_vars = hd.add_vars
            assert add_vars == exp_props['add_vars']



# Copyright 2020,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module for obtaining debug information.

The debug information can be used when reporting bugs to the zhmcclient issue
tracker.

Programmatic usage:
    import zhmcclient
    print(zhmcclient.debuginfo())

Command line usage:
    python -c "import zhmcclient; print(zhmcclient.debuginfo())"
"""


import sys
import platform
import ctypes

from ._version import __version__

__all__ = ['debuginfo']


def _version_string(version_info):
    """
    Return the 5-tuple version_info as a version string, as follows:

        "1.2.3"  # if version_info[3] == 'final'
        "1.2.3.alpha.42"  # if version_info[3] != 'final'
    """
    major, minor, micro, releaselevel, serial = version_info
    version_str = f'{major}.{minor}.{micro}'
    if releaselevel != 'final':
        version_str = f'{version_str}.{releaselevel}.{serial}'
    return version_str


def debuginfo():
    """
    Return debug information as a multi-line string.
    """
    di_dict = debuginfo_dict()
    ret = ""
    for k, v in di_dict.items():
        ret += f"{k}: {v}\n"
    return ret


def debuginfo_dict():
    """
    Return debug information as a dictionary.
    """

    debug_info = {}

    debug_info['os_name'] = platform.system()
    debug_info['os_version'] = platform.release()

    debug_info['cpu_arch'] = platform.machine()
    debug_info['bit_size'] = ctypes.sizeof(ctypes.c_void_p) * 8

    char_len = len(b'\\U00010142'.decode('unicode-escape'))
    if char_len == 1:
        debug_info['unicode_size'] = 'wide'
    elif char_len == 2:
        debug_info['unicode_size'] = 'narrow'
    else:
        # Should not happen
        debug_info['unicode_size'] = 'unknown'

    impl = platform.python_implementation()
    debug_info['python_impl'] = impl

    if impl == 'CPython':
        impl_version = _version_string(sys.version_info)
    elif impl == 'PyPy':
        # pylint: disable=no-member
        impl_version = _version_string(sys.pypy_version_info)
    elif impl == 'Jython':
        impl_version = _version_string(sys.version_info)  # TODO: Verify
    elif impl == 'IronPython':
        impl_version = _version_string(sys.version_info)  # TODO: Verify
    else:
        impl_version = 'unknown'
    debug_info['python_impl_version'] = impl_version

    debug_info['python_version'] = platform.python_version()

    debug_info['zhmcclient_version'] = __version__

    return debug_info



# Copyright 2017,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _user_role module.
"""


import re
import copy
import pytest

from zhmcclient import Client, HTTPError, NotFound, BaseResource, UserRole
from zhmcclient_mock import FakedSession
from tests.common.utils import assert_resources


class TestUserRole:
    """All tests for the UserRole and UserRoleManager classes."""

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.

        Set up a faked session, and add a faked Console without any
        child resources.
        """
        # pylint: disable=attribute-defined-outside-init

        self.session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.client = Client(self.session)

        self.faked_console = self.session.hmc.consoles.add({
            'object-id': None,
            # object-uri will be automatically set
            'parent': None,
            'class': 'console',
            'name': 'fake-console1',
            'description': 'Console #1',
        })
        self.console = self.client.consoles.find(name=self.faked_console.name)

    def add_user_role(self, name, type_):
        """
        Add a faked user role object to the faked Console and return it.
        """
        faked_user_role = self.faked_console.user_roles.add({
            'object-id': f'oid-{name}',
            # object-uri will be automatically set
            'parent': '/api/console',
            'class': 'user-role',
            'name': name,
            'description': f'User Role {name}',
            'type': type_,
        })
        return faked_user_role

    def test_user_role_manager_repr(self):
        """Test UserRoleManager.__repr__()."""

        user_role_mgr = self.console.user_roles

        # Execute the code to be tested
        repr_str = repr(user_role_mgr)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf'^{user_role_mgr.__class__.__name__}\s+at\s+'
            rf'0x{id(user_role_mgr):08x}\s+\(\\n.*',
            repr_str)

    def test_user_role_manager_initial_attrs(self):
        """Test initial attributes of UserRoleManager."""

        user_role_mgr = self.console.user_roles

        # Verify all public properties of the manager object
        assert user_role_mgr.resource_class == UserRole
        assert user_role_mgr.class_name == 'user-role'
        assert user_role_mgr.session is self.session
        assert user_role_mgr.parent is self.console
        assert user_role_mgr.console is self.console

    @pytest.mark.parametrize(
        "full_properties_kwargs, prop_names", [
            (dict(full_properties=False),
             ['object-uri', 'name', 'type']),
            (dict(full_properties=True),
             ['object-uri', 'name', 'type', 'description']),
            ({},  # test default for full_properties (False)
             ['object-uri', 'name', 'type']),
        ]
    )
    @pytest.mark.parametrize(
        "filter_args, exp_names", [
            (None,
             ['a', 'b']),
            ({},
             ['a', 'b']),
            ({'name': 'a'},
             ['a']),
            ({'name': 'A'},  # user roles have case-insensitive names
             ['a']),
        ]
    )
    def test_user_role_manager_list(
            self, filter_args, exp_names, full_properties_kwargs, prop_names):
        """Test UserRoleManager.list()."""

        faked_user_role1 = self.add_user_role(name='a', type_='user-defined')
        faked_user_role2 = self.add_user_role(name='b', type_='user-defined')
        faked_user_roles = [faked_user_role1, faked_user_role2]
        exp_faked_user_roles = [u for u in faked_user_roles
                                if u.name in exp_names]
        user_role_mgr = self.console.user_roles

        # Execute the code to be tested
        user_roles = user_role_mgr.list(filter_args=filter_args,
                                        **full_properties_kwargs)

        assert_resources(user_roles, exp_faked_user_roles, prop_names)

    @pytest.mark.parametrize(
        "input_props, exp_prop_names, exp_exc", [
            ({},  # name missing
             None,
             HTTPError({'http-status': 400, 'reason': 5})),
            ({'description': 'fake description X'},  # name missing
             None,
             HTTPError({'http-status': 400, 'reason': 5})),
            ({'name': 'fake-name-x',
              'type': 'user-defined'},  # type not allowed
             None,
             HTTPError({'http-status': 400, 'reason': 6})),
            ({'name': 'fake-name-x',
              'type': 'system-defined'},  # type not allowed
             None,
             HTTPError({'http-status': 400, 'reason': 6})),
            ({'name': 'fake-name-x'},
             ['object-uri', 'name'],
             None),
            ({'name': 'fake-name-x',
              'description': 'fake description X'},
             ['object-uri', 'name', 'description'],
             None),
        ]
    )
    def test_user_role_manager_create(self, input_props, exp_prop_names,
                                      exp_exc):
        """Test UserRoleManager.create()."""

        user_role_mgr = self.console.user_roles

        # Add default system-defined user role
        self.add_user_role(name='hmc-operator-tasks', type_='system-defined')

        if exp_exc is not None:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                user_role_mgr.create(properties=input_props)

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

        else:

            # Execute the code to be tested.
            user_role = user_role_mgr.create(properties=input_props)

            # Check the resource for consistency within itself
            assert isinstance(user_role, UserRole)
            user_role_name = user_role.name
            exp_user_role_name = user_role.properties['name']
            assert user_role_name == exp_user_role_name
            user_role_uri = user_role.uri
            exp_user_role_uri = user_role.properties['object-uri']
            assert user_role_uri == exp_user_role_uri

            # Check the properties against the expected names and values
            for prop_name in exp_prop_names:
                assert prop_name in user_role.properties
                if prop_name in input_props:
                    value = user_role.properties[prop_name]
                    exp_value = input_props[prop_name]
                    assert value == exp_value

    def test_user_role_repr(self):
        """Test UserRole.__repr__()."""

        faked_user_role1 = self.add_user_role(name='a', type_='user-defined')
        user_role1 = self.console.user_roles.find(name=faked_user_role1.name)

        # Execute the code to be tested
        repr_str = repr(user_role1)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf'^{user_role1.__class__.__name__}\s+at\s+'
            rf'0x{id(user_role1):08x}\s+\(\\n.*',
            repr_str)

    @pytest.mark.parametrize(
        "input_name, input_type, exp_exc", [
            ('a', 'user-defined', None),
            # ('b', 'system-defined',
            #  HTTPError({'http-status': 400, 'reason': 312})),
            # TODO: Re-enable once rejection for system-defined roles supported
        ]
    )
    def test_user_role_delete(self, input_name, input_type, exp_exc):
        """Test UserRole.delete()."""

        faked_user_role = self.add_user_role(name=input_name, type_=input_type)
        user_role_mgr = self.console.user_roles
        user_role = user_role_mgr.find(name=faked_user_role.name)

        if exp_exc is not None:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                user_role.delete()

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

            # Check that the user role still exists
            user_role_mgr.find(name=faked_user_role.name)

        else:

            # Execute the code to be tested.
            user_role.delete()

            # Check that the user role no longer exists
            with pytest.raises(NotFound) as exc_info:
                user_role_mgr.find(name=faked_user_role.name)

    def test_user_role_delete_create_same_name(self):
        """Test UserRole.delete() followed by create() with same name."""

        user_role_name = 'faked_a'

        # Add default system-defined user role
        self.add_user_role(name='hmc-operator-tasks', type_='system-defined')

        # Add the user role to be tested
        self.add_user_role(name=user_role_name, type_='user-defined')

        # Input properties for a user role with the same name
        sn_user_role_props = {
            'name': user_role_name,
            'description': 'User Role with same name',
        }

        user_role_mgr = self.console.user_roles
        user_role = user_role_mgr.find(name=user_role_name)

        # Execute the deletion code to be tested
        user_role.delete()

        # Check that the user role no longer exists
        with pytest.raises(NotFound):
            user_role_mgr.find(name=user_role_name)

        # Execute the creation code to be tested.
        user_role_mgr.create(sn_user_role_props)

        # Check that the user role exists again under that name
        sn_user_role = user_role_mgr.find(name=user_role_name)
        description = sn_user_role.get_property('description')
        assert description == sn_user_role_props['description']

    @pytest.mark.parametrize(
        "input_props", [
            {},
            {'description': 'New user role description'},
        ]
    )
    def test_user_role_update_properties(self, input_props):
        """Test UserRole.update_properties()."""

        # Add the user role to be tested
        faked_user_role = self.add_user_role(name='a', type_='user-defined')

        user_role_mgr = self.console.user_roles
        user_role = user_role_mgr.find(name=faked_user_role.name)

        user_role.pull_full_properties()
        saved_properties = copy.deepcopy(user_role.properties)

        # Execute the code to be tested
        user_role.update_properties(properties=input_props)

        # Verify that the resource object already reflects the property
        # updates.
        for prop_name in saved_properties:
            if prop_name in input_props:
                exp_prop_value = input_props[prop_name]
            else:
                exp_prop_value = saved_properties[prop_name]
            assert prop_name in user_role.properties
            prop_value = user_role.properties[prop_name]
            assert prop_value == exp_prop_value, \
                f"Unexpected value for property {prop_name!r}"

        # Refresh the resource object and verify that the resource object
        # still reflects the property updates.
        user_role.pull_full_properties()
        for prop_name in saved_properties:
            if prop_name in input_props:
                exp_prop_value = input_props[prop_name]
            else:
                exp_prop_value = saved_properties[prop_name]
            assert prop_name in user_role.properties
            prop_value = user_role.properties[prop_name]
            assert prop_value == exp_prop_value

    @pytest.mark.parametrize(
        "role_name, role_type, exp_exc", [
            ('ra', 'user-defined', None),
            ('rb', 'system-defined',
             HTTPError({'http-status': 400, 'reason': 314})),
        ]
    )
    @pytest.mark.parametrize(
        "perm_args", [
            {'permitted_object': 'cpc',
             'include_members': True,
             'view_only': False},
        ]
    )
    def test_user_role_add_permission(
            self, perm_args, role_name, role_type, exp_exc):
        """Test UserRole.add_permission()."""

        faked_user_role = self.add_user_role(name=role_name, type_=role_type)
        user_role_mgr = self.console.user_roles
        user_role = user_role_mgr.find(name=faked_user_role.name)

        permitted_object = perm_args['permitted_object']
        if isinstance(permitted_object, BaseResource):
            perm_obj = permitted_object.uri
            perm_type = 'object'
        else:
            assert isinstance(permitted_object, str)
            perm_obj = permitted_object
            perm_type = 'object-class'
        permission_parms = {
            'permitted-object': perm_obj,
            'permitted-object-type': perm_type,
            'include-members': perm_args['include_members'],
            'view-only-mode': perm_args['view_only'],
        }

        if exp_exc is not None:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                user_role.add_permission(**perm_args)

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

            # Check that the user role still does not have that permission
            user_role.pull_full_properties()
            if 'permissions' in user_role.properties:
                permissions = user_role.properties['permissions']
                assert permission_parms not in permissions

        else:

            # Execute the code to be tested.
            ret = user_role.add_permission(**perm_args)

            assert ret is None

            # Check that the user role now has that permission
            user_role.pull_full_properties()
            permissions = user_role.properties['permissions']
            assert permission_parms in permissions

    @pytest.mark.parametrize(
        "role_name, role_type, exp_exc", [
            ('ra', 'user-defined', None),
            ('rb', 'system-defined',
             HTTPError({'http-status': 400, 'reason': 314})),
        ]
    )
    @pytest.mark.parametrize(
        "perm_args", [
            {'permitted_object': 'cpc',
             'include_members': True,
             'view_only': False},
        ]
    )
    def test_user_role_remove_permission(
            self, perm_args, role_name, role_type, exp_exc):
        """Test UserRole.remove_permission()."""

        faked_user_role = self.add_user_role(name=role_name, type_=role_type)
        user_role_mgr = self.console.user_roles
        user_role = user_role_mgr.find(name=faked_user_role.name)

        permitted_object = perm_args['permitted_object']
        if isinstance(permitted_object, BaseResource):
            perm_obj = permitted_object.uri
            perm_type = 'object'
        else:
            assert isinstance(permitted_object, str)
            perm_obj = permitted_object
            perm_type = 'object-class'
        permission_parms = {
            'permitted-object': perm_obj,
            'permitted-object-type': perm_type,
            'include-members': perm_args['include_members'],
            'view-only-mode': perm_args['view_only'],
        }

        # Prepare the user role with the initial permission
        if 'permissions' not in faked_user_role.properties:
            faked_user_role.properties['permissions'] = []
        faked_user_role.properties['permissions'].append(permission_parms)

        if exp_exc is not None:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                user_role.remove_permission(**perm_args)

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

            # Check that the user role still has that permission
            user_role.pull_full_properties()
            permissions = user_role.properties['permissions']
            assert permission_parms in permissions

        else:

            # Execute the code to be tested.
            ret = user_role.remove_permission(**perm_args)

            assert ret is None

            # Check that the user role no longer has that permission
            user_role.pull_full_properties()
            if 'permissions' in user_role.properties:
                permissions = user_role.properties['permissions']
                assert permission_parms not in permissions



#!/usr/bin/env python
# Copyright 2021,2022 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Example that exports a DPM configuration from a CPC and checks it for
consistency with the inventory data.
"""

import sys
import requests.packages.urllib3

import zhmcclient
from zhmcclient.testutils import hmc_definitions

requests.packages.urllib3.disable_warnings()

# Get HMC info from HMC inventory and vault files
hmc_def = hmc_definitions()[0]
nickname = hmc_def.nickname
host = hmc_def.host
userid = hmc_def.userid
password = hmc_def.password
verify_cert = hmc_def.verify_cert


def check(config_items, inventory_data, classname, uri_prop, prop, value):
    """Check an item in the exported DPM config against the inventory data."""
    if config_items is None:
        return
    if not isinstance(value, (list, tuple)):
        value = [value]
    inventory_items = [x for x in inventory_data if x['class'] == classname
                       and (True if prop is None else x[prop] in value)]
    extra_uris = {x[uri_prop] for x in config_items} - \
        {x[uri_prop] for x in inventory_items}
    missing_uris = {x[uri_prop] for x in inventory_items} - \
        {x[uri_prop] for x in config_items}
    status = 'Error' if extra_uris or missing_uris else 'Ok'
    print(f"Checking resource class {classname!r}: "
          f"Inventory: {len(inventory_items)}; DPM config: {len(config_items)} "
          f"(extra: {len(extra_uris)}, missing: {len(missing_uris)}); "
          f"Check status: {status}")


print(__doc__)

print(f"Using HMC {nickname} at {host} with userid {userid} ...")

print("Creating a session with the HMC ...")
try:
    session = zhmcclient.Session(
        host, userid, password, verify_cert=verify_cert)
except zhmcclient.Error as exc:
    print(f"Error: Cannot establish session with HMC {host}: "
          f"{exc.__class__.__name__}: {exc}")
    sys.exit(1)

try:
    client = zhmcclient.Client(session)

    print("Finding a CPC in DPM mode ...")
    cpcs = client.cpcs.list(filter_args={'dpm-enabled': True})
    if not cpcs:
        print(f"Error: HMC at {host} does not manage any CPCs in DPM mode")
        sys.exit(1)
    cpc = cpcs[0]
    print(f"Using CPC {cpc.name}")

    print(f"Exporting DPM configuration of CPC {cpc.name} ...")
    try:
        dpm_config = cpc.export_dpm_configuration()
    except zhmcclient.ConsistencyError as exc:
        print(f"Error: Cannot export DPM configuration: {exc}")
        sys.exit(1)
    fields_str = ', '.join(dpm_config.keys())
    print(f"Fields in exported DPM configuration: {fields_str}")

    print("Checking some items in the exported DPM configuration for "
          "consistency with the inventory data ...")

    print("Retrieving inventory data ...")
    inventory_data = client.get_inventory(['dpm-resources', 'cpc'])

    adapter_uris = [x['object-uri'] for x in inventory_data
                    if x['class'] == 'adapter' and x['parent'] == cpc.uri]
    partition_uris = [x['object-uri'] for x in inventory_data
                      if x['class'] == 'partition' and x['parent'] == cpc.uri]
    storage_group_uris = [x['object-uri'] for x in inventory_data
                          if x['class'] == 'storage-group'
                          and x['cpc-uri'] == cpc.uri]

    check(dpm_config.get('adapters'), inventory_data,
          'adapter', 'object-uri', 'parent', cpc.uri)
    check(dpm_config.get('network-ports'), inventory_data,
          'network-port', 'element-uri', 'parent', adapter_uris)
    check(dpm_config.get('storage-ports'), inventory_data,
          'storage-port', 'element-uri', 'parent', adapter_uris)

    check(dpm_config.get('partitions'), inventory_data,
          'partition', 'object-uri', 'parent', cpc.uri)
    check(dpm_config.get('nics'), inventory_data,
          'nic', 'element-uri', 'parent', partition_uris)
    check(dpm_config.get('hbas'), inventory_data,
          'hba', 'element-uri', 'parent', partition_uris)
    check(dpm_config.get('virtual-functions'), inventory_data,
          'virtual-function', 'element-uri', 'parent', partition_uris)

    check(dpm_config.get('virtual-switches'), inventory_data,
          'virtual-switch', 'object-uri', 'parent', cpc.uri)
    check(dpm_config.get('capacity-groups'), inventory_data,
          'capacity-group', 'element-uri', 'parent', cpc.uri)

    check(dpm_config.get('storage-sites'), inventory_data,
          'storage-site', 'object-uri', None, None)

    check(dpm_config.get('storage-groups'), inventory_data,
          'storage-group', 'object-uri', 'cpc-uri', cpc.uri)
    check(dpm_config.get('storage-volumes'), inventory_data,
          'storage-volume', 'element-uri', 'parent', storage_group_uris)

finally:
    print("Logging off ...")
    session.logoff()



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _session module.
"""


import time
import json
import re
from unittest import mock
import requests
import requests_mock
import pytest

from zhmcclient import Session, ParseError, Job, HTTPError, OperationTimeout, \
    ClientAuthError, DEFAULT_HMC_PORT

# Default value for the 'verify_cert' parameter of the Session class:
DEFAULT_VERIFY_CERT = True


# TODO: Test Session.get() in all variations (including errors)
# TODO: Test Session.post() in all variations (including errors)
# TODO: Test Session.delete() in all variations (including errors)

def mock_server_1(m):
    """
    Set up the mocked responses for a simple HMC server that supports
    logon and logoff.
    """
    m.register_uri('POST', '/api/sessions',
                   json={
                       'api-session': 'test-session-id',
                       'notification-topic': 'test-obj-topic.1',
                       'job-notification-topic': 'test-job-topic.1',
                       'session-credential':
                           'un8bu462g37aw9j0o8pltontz3szt35jh4b1qe2toxt6fkhl4',
                   },
                   headers={'X-Request-Id': 'fake-request-id'})
    m.register_uri('DELETE', '/api/sessions/this-session',
                   status_code=204)
    m.register_uri('GET', '/api/version',
                   json={
                       'api-major-version': 4,
                       'api-minor-version': 10,
                   })


@pytest.mark.parametrize(
    "host, userid, password, use_get_password, session_id, kwargs", [
        ('fake-host', None, None, False, None, {}),
        ('fake-host', 'fake-userid', None, False, None, {}),
        ('fake-host', 'fake-userid', 'fake-pw', False, None, {}),
        ('fake-host', 'fake-userid', 'fake-pw', True, None, {}),
        ('fake-host', 'fake-userid', 'fake-pw', True, None,
         {'port': 1234}),
        ('fake-host', 'fake-userid', 'fake-pw', True, None,
         {'verify_cert': True}),
    ]
)
def test_session_init(
        host, userid, password, use_get_password, session_id, kwargs):
    """Test initialization of Session object."""

    # TODO: Add support for input parameter: retry_timeout_config
    # TODO: Add support for input parameter: time_stats_keeper

    if use_get_password:
        def get_password(host, userid):
            pw = f'fake-pw-{host}-{userid}'
            return pw
    else:
        get_password = None

    session = Session(host, userid, password, session_id, get_password,
                      **kwargs)

    assert session.host == host
    assert session.userid == userid
    # pylint: disable=protected-access
    assert session._password == password
    assert session.session_id == session_id
    assert session.get_password == get_password
    assert session.port == kwargs.get('port', DEFAULT_HMC_PORT)
    assert session.verify_cert == kwargs.get('verify_cert', DEFAULT_VERIFY_CERT)

    assert session.headers['Content-type'] == 'application/json'
    assert session.headers['Accept'] == '*/*'

    if session_id is None:
        assert session.session is None
        assert 'X-API-Session' not in session.headers
        assert len(session.headers) == 3
        assert session.actual_host is None
        assert session.base_url is None
    else:
        assert isinstance(session.session, requests.Session)
        assert session.headers['X-API-Session'] == session_id
        assert len(session.headers) == 4
        assert session.actual_host == host
        base_url = f'https://{session.actual_host}:{session.port!s}'
        assert session.base_url == base_url


def test_session_repr():
    """Test Session.__repr__()."""

    session = Session('fake-host', 'fake-user', 'fake-pw')

    repr_str = repr(session)

    repr_str = repr_str.replace('\n', '\\n')
    # We check just the begin of the string:
    assert re.match(
        rf'^{session.__class__.__name__}\s+at\s+'
        rf'0x{id(session):08x}\s+\(\\n.*',
        repr_str)


@pytest.mark.parametrize(
    "host, userid, password, use_get_password, exp_exc", [
        ('fake-host', None, None, False, ClientAuthError),
        ('fake-host', 'fake-userid', None, False, ClientAuthError),
        ('fake-host', 'fake-userid', 'fake-pw', False, None),
        ('fake-host', 'fake-userid', 'fake-pw', True, None),
        ('fake-host', 'fake-userid', None, True, None),
    ]
)
def test_session_logon(
        host, userid, password, use_get_password, exp_exc):
    """Test Session.logon() (and also Session.is_logon())."""

    with requests_mock.Mocker() as m:

        mock_server_1(m)

        if use_get_password:
            get_password = mock.MagicMock()
            get_password.return_value = \
                f'fake-pw-{host}-{userid}'
        else:
            get_password = None

        # Create a session in logged-off state
        session = Session(host, userid, password, None, get_password)

        assert session.session_id is None
        assert 'X-API-Session' not in session.headers
        assert session.session is None

        logged_on = session.is_logon()
        assert not logged_on

        if exp_exc:
            try:

                # The code to be tested:
                session.logon()

            except exp_exc:
                pass

            logged_on = session.is_logon()
            assert not logged_on
        else:

            # The code to be tested:
            session.logon()

            assert session.session_id == 'test-session-id'
            assert 'X-API-Session' in session.headers
            assert isinstance(session.session, requests.Session)

            if get_password:
                if password is None:
                    get_password.assert_called_with(host, userid)
                    # pylint: disable=protected-access
                    assert session._password == get_password.return_value
                else:
                    get_password.assert_not_called()

            logged_on = session.is_logon()
            assert logged_on


def test_session_logoff():
    """Test Session.logoff() (and also Session.is_logon())."""

    with requests_mock.Mocker() as m:

        mock_server_1(m)

        # Create a session in logged-off state
        session = Session('fake-host', 'fake-userid', 'fake-pw')

        session.logon()

        logged_on = session.is_logon()
        assert logged_on

        # The code to be tested:
        session.logoff()

        assert session.session_id is None
        assert session.session is None
        assert 'X-API-Session' not in session.headers
        assert len(session.headers) == 3

        logged_on = session.is_logon()
        assert not logged_on


def _do_parse_error_logon(m, json_content, exp_msg_pattern, exp_line, exp_col):
    """
    Perform a session logon, and mock the provided (invalid) JSON content
    for the response so that a JSON parsing error is triggered.

    Assert that this is surfaced via a `zhmcclient.ParseError` exception,
    with the expected message (as a regexp pattern), line and column.
    """

    m.register_uri('POST', '/api/sessions',
                   content=json_content,
                   headers={'X-Request-Id': 'fake-request-id'})
    m.register_uri('GET', '/api/version',
                   json={
                       'api-major-version': 4,
                       'api-minor-version': 10,
                   })

    session = Session('fake-host', 'fake-user', 'fake-pw')

    exp_pe_pattern = \
        rf"^JSON parse error in HTTP response: {exp_msg_pattern}\. " \
        r"HTTP request: [^ ]+ [^ ]+\. " \
        r"Response status .*"

    with pytest.raises(ParseError) as exc_info:
        session.logon()
    exc = exc_info.value

    assert re.match(exp_pe_pattern, str(exc))
    assert exc.line == exp_line
    assert exc.column == exp_col


# TODO: Merge the next 3 test functions into one that is parametrized

@requests_mock.mock()
def test_session_logon_error_invalid_delim(*args):
    """
    Logon with invalid JSON response that has an invalid delimiter.
    """
    m = args[0]
    json_content = b'{\n"api-session"; "test-session-id"\n}'
    exp_msg_pattern = r"Expecting ':' delimiter: .*"
    exp_line = 2
    exp_col = 14
    _do_parse_error_logon(m, json_content, exp_msg_pattern, exp_line, exp_col)


@requests_mock.mock()
def test_session_logon_error_invalid_quotes(*args):
    """
    Logon with invalid JSON response that incorrectly uses single quotes.
    """
    m = args[0]
    json_content = b'{\'api-session\': \'test-session-id\'}'
    exp_msg_pattern = r"Expecting property name enclosed in double " \
        "quotes: .*"
    exp_line = 1
    exp_col = 2
    _do_parse_error_logon(m, json_content, exp_msg_pattern, exp_line, exp_col)


@requests_mock.mock()
def test_session_logon_error_extra_closing(*args):
    """
    Logon with invalid JSON response that has an extra closing brace.
    """
    m = args[0]
    json_content = b'{"api-session": "test-session-id"}}'
    exp_msg_pattern = r"Extra data: .*"
    exp_line = 1
    exp_col = 35
    _do_parse_error_logon(m, json_content, exp_msg_pattern, exp_line, exp_col)


def test_session_get_notification_topics():
    """
    This tests the 'Get Notification Topics' operation.
    """
    session = Session('fake-host', 'fake-user', 'fake-id')
    with requests_mock.mock() as m:
        # Because logon is deferred until needed, we perform it
        # explicitly in order to keep mocking in the actual test simple.
        m.post(
            '/api/sessions', json={
                'api-session': 'test-session-id',
                'notification-topic': 'test-obj-topic.1',
                'job-notification-topic': 'test-job-topic.1',
                'session-credential':
                    'un8bu462g37aw9j0o8pltontz3szt35jh4b1qe2toxt6fkhl4',
            })
        m.get(
            '/api/version', json={
                'api-major-version': 4,
                'api-minor-version': 10,
            })
        session.logon()
        gnt_uri = "/api/sessions/operations/get-notification-topics"
        gnt_result = {
            "topics": [
                {
                    'topic-name': 'ensadmin.145',
                    'topic-type': 'object-notification',
                },
                {
                    'topic-name': 'ensadmin.145job',
                    'topic-type': 'job-notification',
                },
                {
                    'topic-name': 'ensadmin.145aud',
                    'topic-type': 'audit-notification',
                },
                {
                    'topic-name': 'ensadmin.145sec',
                    'topic-type': 'security-notification',
                }
            ]
        }
        m.get(gnt_uri, json=gnt_result)

        result = session.get_notification_topics()

        assert result == gnt_result['topics']

        m.delete('/api/sessions/this-session', status_code=204)

        session.logoff()


def test_session_get_error_html_1():
    """
    This tests a dummy GET with a 500 response with HTML content.
    """
    session = Session('fake-host', 'fake-user', 'fake-id')
    with requests_mock.mock() as m:
        get_uri = "/api/version"
        get_resp_status = 500
        get_resp_content_type = 'text/html; charset=utf-8'
        get_resp_headers = {
            'content-type': get_resp_content_type,
        }
        get_resp_content = """\
<!doctype html public "-//IETF//DTD HTML 2.0//EN">\
 <html>\
<head>\
<title>Console Internal Error</title>\
 <link href="/skin/HMCskin.css" rel="stylesheet" type="text/css"/>\
</head>\
 <body>\
<h1>Console Internal Error</h1>\
<br><hr size="1" noshade>\
<h2>Details:</h2>\
<p><br>HTTP status code: 500\
<p><br>The server encountered an internal error that prevented it from\
 fulfilling this request.\
<p><br>\
<pre>javax.servlet.ServletException: Web Services are not enabled.
\tat com.ibm.hwmca.fw.api.ApiServlet.execute(ApiServlet.java:135)
\t. . .
</pre>\
<hr size="1" noshade>\
</body>\
</html>"""
        m.get(get_uri, text=get_resp_content, headers=get_resp_headers,
              status_code=get_resp_status)

        # The following expected results reflect what is done in
        # _session._result_object().

        exp_reason = 900
        exp_message = \
            "Console Configuration Error: " \
            "Web Services API is not enabled on the HMC."

        with pytest.raises(HTTPError) as exc_info:
            session.get(get_uri, logon_required=False)
        exc = exc_info.value

        assert exc.http_status == get_resp_status
        assert exc.reason == exp_reason
        assert exc.message == exp_message
        assert exc.request_uri.endswith(get_uri)
        assert exc.request_method == 'GET'


JOB_URI = '/api/jobs/fake-job-uri'


# TODO: Add parametrization to the next test function.

def test_job_init():
    """Test initialization of Job object."""
    session = Session('fake-host', 'fake-user', 'fake-pw')

    # Jobs exist only for POST, but we want to test that the specified HTTP
    # method comes back regardless:
    op_method = 'GET'

    op_uri = '/api/bla'

    job = Job(session, JOB_URI, op_method, op_uri)

    assert job.uri == JOB_URI
    assert job.session == session
    assert job.op_method == op_method
    assert job.op_uri == op_uri


# TODO: Merge the next 7 test functions into one that is parametrized

def test_job_check_incomplete():
    """Test check_for_completion() with incomplete job."""
    with requests_mock.mock() as m:
        mock_server_1(m)
        session = Session('fake-host', 'fake-user', 'fake-pw')
        op_method = 'POST'
        op_uri = '/api/foo'
        job = Job(session, JOB_URI, op_method, op_uri)
        query_job_status_result = {
            'status': 'running',
        }
        m.get(JOB_URI, json=query_job_status_result)
        m.delete(JOB_URI, status_code=204)

        job_status, op_result = job.check_for_completion()

        assert job_status == 'running'
        assert op_result is None


def test_job_check_complete_success_noresult():
    """Test check_for_completion() with successful complete job without
    result."""
    with requests_mock.mock() as m:
        mock_server_1(m)
        session = Session('fake-host', 'fake-user', 'fake-pw')
        op_method = 'POST'
        op_uri = '/api/foo'
        job = Job(session, JOB_URI, op_method, op_uri)
        query_job_status_result = {
            'status': 'complete',
            'job-status-code': 200,
            # 'job-reason-code' omitted because HTTP status good
            # 'job-results' is optional and is omitted
        }
        m.get(JOB_URI, json=query_job_status_result)
        m.delete(JOB_URI, status_code=204)

        job_status, op_result = job.check_for_completion()

        assert job_status == 'complete'
        assert op_result is None


def test_job_check_complete_success_result():
    """Test check_for_completion() with successful complete job with a
    result."""
    with requests_mock.mock() as m:
        mock_server_1(m)
        session = Session('fake-host', 'fake-user', 'fake-pw')
        op_method = 'POST'
        op_uri = '/api/foo'
        job = Job(session, JOB_URI, op_method, op_uri)
        exp_op_result = {
            'foo': 'bar',
        }
        query_job_status_result = {
            'status': 'complete',
            'job-status-code': 200,
            # 'job-reason-code' omitted because HTTP status good
            'job-results': exp_op_result,
        }
        m.get(JOB_URI, json=query_job_status_result)
        m.delete(JOB_URI, status_code=204)

        job_status, op_result = job.check_for_completion()

        assert job_status == 'complete'
        assert op_result == exp_op_result


def test_job_check_complete_error1():
    """Test check_for_completion() with complete job in error (1)."""
    with requests_mock.mock() as m:
        mock_server_1(m)
        session = Session('fake-host', 'fake-user', 'fake-pw')
        op_method = 'POST'
        op_uri = '/api/foo'
        job = Job(session, JOB_URI, op_method, op_uri)
        query_job_status_result = {
            'status': 'complete',
            'job-status-code': 500,
            'job-reason-code': 42,
            # no 'job-results' field (it is not guaranteed to be there)
        }

        m.get(JOB_URI, json=query_job_status_result)
        m.delete(JOB_URI, status_code=204)

        with pytest.raises(HTTPError) as exc_info:
            job.check_for_completion()
        exc = exc_info.value

        assert exc.http_status == 500
        assert exc.reason == 42
        assert exc.message is None


def test_job_check_complete_error2():
    """Test check_for_completion() with complete job in error (2)."""
    with requests_mock.mock() as m:
        mock_server_1(m)
        session = Session('fake-host', 'fake-user', 'fake-pw')
        op_method = 'POST'
        op_uri = '/api/foo'
        job = Job(session, JOB_URI, op_method, op_uri)
        query_job_status_result = {
            'status': 'complete',
            'job-status-code': 500,
            'job-reason-code': 42,
            'job-results': {},  # it is not guaranteed to have any content
        }

        m.get(JOB_URI, json=query_job_status_result)
        m.delete(JOB_URI, status_code=204)

        with pytest.raises(HTTPError) as exc_info:
            job.check_for_completion()
        exc = exc_info.value

        assert exc.http_status == 500
        assert exc.reason == 42
        assert exc.message is None


def test_job_check_complete_error3():
    """Test check_for_completion() with complete job in error (3)."""
    with requests_mock.mock() as m:
        mock_server_1(m)
        session = Session('fake-host', 'fake-user', 'fake-pw')
        op_method = 'POST'
        op_uri = '/api/foo'
        job = Job(session, JOB_URI, op_method, op_uri)
        query_job_status_result = {
            'status': 'complete',
            'job-status-code': 500,
            'job-reason-code': 42,
            'job-results': {
                # Content is not documented for the error case.
                # Some failures result in an 'error' field.
                'error': 'bla message',
            },
        }

        m.get(JOB_URI, json=query_job_status_result)
        m.delete(JOB_URI, status_code=204)

        with pytest.raises(HTTPError) as exc_info:
            job.check_for_completion()
        exc = exc_info.value

        assert exc.http_status == 500
        assert exc.reason == 42
        assert exc.message == 'bla message'


def test_job_check_complete_error4():
    """Test check_for_completion() with complete job in error (4)."""
    with requests_mock.mock() as m:
        mock_server_1(m)
        session = Session('fake-host', 'fake-user', 'fake-pw')
        op_method = 'POST'
        op_uri = '/api/foo'
        job = Job(session, JOB_URI, op_method, op_uri)
        query_job_status_result = {
            'status': 'complete',
            'job-status-code': 500,
            'job-reason-code': 42,
            'job-results': {
                # Content is not documented for the error case.
                # Some failures result in an 'message' field.
                'message': 'bla message',
            },
        }

        m.get(JOB_URI, json=query_job_status_result)
        m.delete(JOB_URI, status_code=204)

        with pytest.raises(HTTPError) as exc_info:
            job.check_for_completion()
        exc = exc_info.value

        assert exc.http_status == 500
        assert exc.reason == 42
        assert exc.message == 'bla message'


# TODO: Merge the next 3 test functions into one that is parametrized

def test_job_wait_complete1_success_result():
    """Test wait_for_completion() with successful complete job with a
    result."""
    with requests_mock.mock() as m:
        mock_server_1(m)
        session = Session('fake-host', 'fake-user', 'fake-pw')
        op_method = 'POST'
        op_uri = '/api/foo'
        job = Job(session, JOB_URI, op_method, op_uri)
        exp_op_result = {
            'foo': 'bar',
        }
        query_job_status_result = {
            'status': 'complete',
            'job-status-code': 200,
            # 'job-reason-code' omitted because HTTP status good
            'job-results': exp_op_result,
        }
        m.get(JOB_URI, json=query_job_status_result)
        m.delete(JOB_URI, status_code=204)

        op_result = job.wait_for_completion()

        assert op_result == exp_op_result


def test_job_wait_complete3_success_result():
    """Test wait_for_completion() with successful complete job with a
    result."""
    with requests_mock.mock() as m:
        mock_server_1(m)
        session = Session('fake-host', 'fake-user', 'fake-pw')
        op_method = 'POST'
        op_uri = '/api/foo'
        job = Job(session, JOB_URI, op_method, op_uri)
        exp_op_result = {
            'foo': 'bar',
        }
        m.get(JOB_URI,
              [
                  {'text': result_running_callback},
                  {'text': result_complete_callback},
              ])
        m.delete(JOB_URI, status_code=204)

        op_result = job.wait_for_completion()

        assert op_result == exp_op_result


def test_job_wait_complete3_timeout():
    """Test wait_for_completion() with timeout."""
    with requests_mock.mock() as m:
        mock_server_1(m)
        session = Session('fake-host', 'fake-user', 'fake-pw')
        op_method = 'POST'
        op_uri = '/api/foo'
        job = Job(session, JOB_URI, op_method, op_uri)
        m.get(JOB_URI,
              [
                  {'text': result_running_callback},
                  {'text': result_running_callback},
                  {'text': result_complete_callback},
              ])
        m.delete(JOB_URI, status_code=204)

        # Here we provoke a timeout, by setting the timeout to less than
        # the time it would take to return the completed job status.
        # The time it would take is the sum of the following:
        # - 2 * 1 s (1 s is the sleep time in Job.wait_for_completion(),
        #   and this happens before each repeated status retrieval)
        # - 3 * 1 s (1 s is the sleep time in result_*_callback() in this
        #   module, and this happens in each mocked status retrieval)
        # Because status completion is given priority over achieving the
        # timeout duration, the timeout value needed to provoke the
        # timeout exception needs to be shorter by the last status
        # retrieval (the one that completes the job), so 3 s is the
        # boundary for the timeout value.
        operation_timeout = 2.9
        try:
            start_time = time.time()
            job.wait_for_completion(operation_timeout=operation_timeout)
            duration = time.time() - start_time
            raise AssertionError(
                f"No OperationTimeout raised. Actual duration: {duration} s, "
                f"timeout: {operation_timeout} s")
        except OperationTimeout as exc:
            msg = exc.args[0]
            assert msg.startswith("Waiting for completion of job")


def result_running_callback(request, context):
    # pylint: disable=unused-argument
    """
    Callback function returning 'running' JSON.
    """
    job_result_running = {
        'status': 'running',
    }
    time.sleep(1)
    return json.dumps(job_result_running)


def result_complete_callback(request, context):
    # pylint: disable=unused-argument
    """
    Callback function returning 'complete' JSON.
    """
    exp_op_result = {
        'foo': 'bar',
    }
    job_result_complete = {
        'status': 'complete',
        'job-status-code': 200,
        # 'job-reason-code' omitted because HTTP status good
        'job-results': exp_op_result,
    }
    time.sleep(1)
    return json.dumps(job_result_complete)



# Copyright 2019,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Pytest fixtures for mocked HMCs.
"""


import os
import logging
import pytest
import zhmcclient
import zhmcclient_mock

from ._hmc_definition import HMCDefinition
from ._hmc_definitions import hmc_definitions

__all__ = ['hmc_definition', 'hmc_session']

# Log file
TESTLOGFILE = os.getenv('TESTLOGFILE', None)
if TESTLOGFILE:
    LOG_HANDLER = logging.FileHandler(TESTLOGFILE, encoding='utf-8')
    LOG_FORMAT_STRING = '%(asctime)s %(name)s %(levelname)s %(message)s'
    LOG_HANDLER.setFormatter(logging.Formatter(LOG_FORMAT_STRING))
else:
    LOG_HANDLER = None
    LOG_FORMAT_STRING = None


def fixtureid_hmc_definition(fixture_value):
    """
    Return a fixture ID to be used by pytest, for fixture `hmc_definition()`.

    Parameters:

      fixture_value (HMCDefinition): The HMC definition of the HMC the test
        runs against.
    """
    if not isinstance(fixture_value, HMCDefinition):
        return None  # Use pytest auto-generated ID
    hd = fixture_value
    show_hmc = False  # Enable to show HMC/userid or mock file in tests
    if not show_hmc:
        hmc_str = ""
    elif hd.mock_file:
        hmc_str = f"(mock_file={hd.mock_file})"
    else:
        hmc_str = f"(host={hd.host}, userid={hd.userid})"
    ret_str = f"hmc_definition={hd.nickname}{hmc_str}"
    return ret_str


@pytest.fixture(
    params=hmc_definitions(load=None),
    scope='module',
    ids=fixtureid_hmc_definition
)
def hmc_definition(request):
    """
    Pytest fixture representing the set of HMC definitions to use for a test.

    A test function parameter with the name of this fixture resolves to the
    :class:`~zhmcclient.testutils.HMCDefinition`
    object of each HMC to test against.

    The test function is called once for each HMC, if the targeted HMC is a
    group.
    """
    return request.param


@pytest.fixture(
    scope='module'
)
def hmc_session(request, hmc_definition):
    # pylint: disable=redefined-outer-name,unused-argument
    """
    Pytest fixture representing the set of HMC sessions to run a test against.

    A test function parameter with the name of this fixture resolves to the
    :class:`zhmcclient.Session` or :class:`zhmcclient_mock.FakedSession` object
    for each HMC to test against.

    The session is already logged on to the HMC.

    The session object has an additional property named ``hmc_definition``
    that is the :class:`~zhmcclient.testutils.HMCDefinition` object for the
    corresponding HMC definition in the :ref:`HMC inventory file`.

    Because the `hmc_definition` parameter of this fixture is again a fixture,
    the :func:`zhmcclient.testutils.hmc_definition` function needs to be
    imported as well when this fixture is used.

    Upon fixture teardown, the session is automatically logged off from the HMC.
    """
    session = setup_hmc_session(hmc_definition)
    yield session
    teardown_hmc_session(session)


def setup_hmc_session(hd):
    """
    Setup an HMC session and return a new session object for it.

    If the HMC definition represents a real HMC, log on to an HMC and return
    a new :class:`zhmcclient.Session` object.

    If the HMC definition represents a mocked HMC, create a new mock environment
    from that and return a :class:`zhmcclient_mock.FakedSession` object.
    """
    # We use the cached skip reason from previous attempts
    skip_msg = getattr(hd, 'skip_msg', None)
    if skip_msg:
        pytest.skip(f"Skip reason from earlier attempt: {skip_msg}")

    if hd.mock_file:
        # A mocked HMC

        # Create a mocked session using the mock file from the inventory file
        session = zhmcclient_mock.FakedSession.from_hmc_yaml_file(
            hd.mock_file, userid=hd.userid, password=hd.password)

        # Set the HMC definition host to the host found in the mock file.
        hd.host = session.host

    else:
        # A real HMC

        # Enable debug logging if specified
        if LOG_HANDLER:

            logger = logging.getLogger('zhmcclient.hmc')
            if LOG_HANDLER not in logger.handlers:
                logger.addHandler(LOG_HANDLER)
            logger.setLevel(logging.DEBUG)

            logger = logging.getLogger('zhmcclient.api')
            if LOG_HANDLER not in logger.handlers:
                logger.addHandler(LOG_HANDLER)
            logger.setLevel(logging.DEBUG)

            logger = logging.getLogger('zhmcclient.jms')
            if LOG_HANDLER not in logger.handlers:
                logger.addHandler(LOG_HANDLER)
            logger.setLevel(logging.DEBUG)

            logger = logging.getLogger('zhmcclient.os')
            if LOG_HANDLER not in logger.handlers:
                logger.addHandler(LOG_HANDLER)
            logger.setLevel(logging.DEBUG)

        rt_config = zhmcclient.RetryTimeoutConfig(
            read_timeout=300,
        )

        # Creating a session does not interact with the HMC (logon is deferred)
        session = zhmcclient.Session(
            hd.host, hd.userid, hd.password, verify_cert=hd.verify_cert,
            retry_timeout_config=rt_config)

        # Check access to the HMC
        try:
            session.logon()
        except zhmcclient.Error as exc:
            msg = (
                f"Cannot log on to HMC {hd.nickname} at {hd.host} "
                f"due to {exc.__class__.__name__}: {exc}")
            hd.skip_msg = msg
            pytest.skip(msg)

    hd.skip_msg = None
    session.hmc_definition = hd

    return session


def teardown_hmc_session(session):
    """
    Log off from an HMC session.
    """
    session.logoff()



#!/usr/bin/env python
# Copyright 2016,2022 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Example that gets the API version of an HMC using an unauthenticated session.
"""

import sys
import requests.packages.urllib3

import zhmcclient
from zhmcclient.testutils import hmc_definitions

requests.packages.urllib3.disable_warnings()

# Get HMC info from HMC inventory and vault files
hmc_def = hmc_definitions()[0]
nickname = hmc_def.nickname
host = hmc_def.host
verify_cert = hmc_def.verify_cert

print(__doc__)

print(f"Using HMC {nickname} at {host} ...")

print("Creating an unauthenticated session with the HMC ...")
try:
    session = zhmcclient.Session(host, verify_cert=verify_cert)
except zhmcclient.Error as exc:
    print(f"Error: Cannot establish session with HMC {host}: "
          f"{exc.__class__.__name__}: {exc}")
    sys.exit(1)

client = zhmcclient.Client(session)

vi = client.version_info()
print(f"HMC API version: {vi[0]}.{vi[1]}")



#!/usr/bin/env python
# Copyright 2016,2022 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Example that performs an asynchronous start of a partition and polls for job
completion.
"""

import sys
import uuid
import requests.packages.urllib3
import time

import zhmcclient
from zhmcclient.testutils import hmc_definitions

requests.packages.urllib3.disable_warnings()

# Get HMC info from HMC inventory and vault files
hmc_def = hmc_definitions()[0]
nickname = hmc_def.nickname
host = hmc_def.host
userid = hmc_def.userid
password = hmc_def.password
verify_cert = hmc_def.verify_cert

print(__doc__)

print(f"Using HMC {nickname} at {host} with userid {userid} ...")

print("Creating a session with the HMC ...")
try:
    session = zhmcclient.Session(
        host, userid, password, verify_cert=verify_cert)
except zhmcclient.Error as exc:
    print(f"Error: Cannot establish session with HMC {host}: "
          f"{exc.__class__.__name__}: {exc}")
    sys.exit(1)

try:
    client = zhmcclient.Client(session)

    print("Finding a CPC in DPM mode ...")
    cpcs = client.cpcs.list(filter_args={'dpm-enabled': True})
    if not cpcs:
        print(f"Error: HMC at {host} does not manage any CPCs in DPM mode")
        sys.exit(1)
    cpc = cpcs[0]
    print(f"Using CPC {cpc.name}")

    part_name = f"zhmc_test_{uuid.uuid4()}"
    print(f"Creating partition {part_name} ...")
    try:
        part = cpc.partitions.create(
            properties={
                'name': part_name,
                'type': 'linux',
                'ifl-processors': 2,
                'initial-memory': 4096,
                'maximum-memory': 4096,
            })
    except zhmcclient.Error as exc:
        print(f"Error: Cannot create partition {part_name} on CPC {cpc.name}: "
              f"{exc.__class__.__name__}: {exc}")
        sys.exit(1)

    try:
        print(f"Starting partition {part.name} asynchronously ...")
        job = part.start(wait_for_completion=False)

        sleep_time = 1
        print(f"Polling for job completion with sleep time {sleep_time} sec ...")
        while True:
            try:
                job_status, op_result = job.check_for_completion()
            except zhmcclient.Error as exc:
                print("Error: Job completed; Start operation failed with "
                      f"{exc.__class__.__name__}: {exc}")
                break
            print(f"Job status: {job_status}")
            if job_status == 'complete':
                break
            time.sleep(sleep_time)
        print("Job completed; Start operation succeeded")

    finally:
        if part.get_property('status') != 'stopped':
            print(f"Stopping partition {part.name} ...")
            try:
                part.stop(wait_for_completion=True)
            except zhmcclient.Error as exc:
                print("Error: Stop operation failed with "
                      f"{exc.__class__.__name__}: {exc}")
                sys.exit(1)

        print(f"Deleting partition {part.name} ...")
        try:
            part.delete()
        except zhmcclient.Error as exc:
            print(f"Error: Cannot delete partition {part.name} on CPC "
                  "{cpc.name} for clean up - "
                  f"Please delete it manually: {exc.__class__.__name__}: {exc}")
            sys.exit(1)

finally:
    print("Logging off ...")
    session.logoff()



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _logging module.
"""


import re
import logging
import pytest
from testfixtures import LogCapture

from zhmcclient._logging import logged_api_call, get_logger
from zhmcclient._constants import BLANKED_OUT_STRING

#
# Various uses of the @logged_api_call decorator
#


@logged_api_call
def decorated_global_function():
    """A decorated function at the global (module) level."""
    pass


@logged_api_call(blanked_properties=['hideme'], properties_pos=0)
def decorated_global_props_function(properties):
    """A decorated function with properties at the global (module) level,
    where the 'hideme' property is blanked out in the API log."""
    return properties


def global1_function():
    """An undecorated function at the global (module) level."""

    @logged_api_call
    def decorated_inner1_function():
        """A decorated inner function defined in a global function."""
        pass

    return decorated_inner1_function


def get_decorated_inner1_function():
    """Return the decorated inner function."""
    return global1_function()


def global2_function():
    """An undecorated function at the global (module) level."""

    def inner1_function():
        """An undecorated function at the inner 1 level."""

        @logged_api_call
        def decorated_inner2_function():
            """A decorated inner function defined in another inner function
            that is defined in a global function."""
            pass

        return decorated_inner2_function

    return inner1_function()


def get_decorated_inner2_function():
    """Return the decorated inner 2 function."""
    return global2_function()


class Decorator1Class:
    # pylint: disable=too-few-public-methods
    """A class that has a decorated method."""

    @logged_api_call
    def decorated_method(self):
        """A decorated method of a class."""
        pass


class Decorator2Class:
    """A class that has a decorated method inside a method."""

    @staticmethod
    def method():
        """A method."""

        @logged_api_call
        def decorated_inner_function():
            """A decorated inner function defined in a method of a class."""
            pass

        return decorated_inner_function

    @staticmethod
    def get_decorated_inner_function():
        """Return the decorated inner function."""
        return Decorator2Class.method()


#
# Supporting definitions
#

class CallerClass:
    # pylint: disable=too-few-public-methods
    """
    A supporting class.
    """

    @staticmethod
    def call_from_method(func, *args, **kwargs):
        """
        A supporting method that calls the specified function with the
        specified arguments and keyword arguments. This is used by the test
        cases so that this function acts as a caller for the decorated API
        function.
        """
        return func(*args, **kwargs)


def call_from_global(func, *args, **kwargs):
    """
    A supporting global function that calls the specified function with the
    specified arguments and keyword arguments. This is used by the test cases
    so that this function acts as a caller for the decorated API function.
    """
    return func(*args, **kwargs)


# Some expected values that are constant
_EXP_LOGGER_NAME = 'zhmcclient.api'
_EXP_LOG_LEVEL = 'DEBUG'
_EXP_LOG_MSG_ENTER_PATTERN = "Called: (.*), args: (.*), kwargs: (.*)"
_EXP_LOG_MSG_LEAVE_PATTERN = "Return: (.*), result: (.*)"


@pytest.fixture()
def capture():
    """
    This way of defining a fixture works around the issue that when
    using the decorator testfixtures.log_capture() instead, pytest
    fails with "fixture 'capture' not found".
    """
    with LogCapture(level=logging.DEBUG) as log:
        yield log


#
# Test cases
#

def assert_log_capture(
        log_capture, *, func_pattern=None, args_pattern=None,
        kwargs_pattern=None, return_pattern=None):
    """
    Assert that the log capture is as expected.
    """
    assert len(log_capture.records) == 2

    enter_record = log_capture.records[0]
    assert enter_record.name == _EXP_LOGGER_NAME
    assert enter_record.levelname == _EXP_LOG_LEVEL
    assert re.match(_EXP_LOG_MSG_ENTER_PATTERN, enter_record.msg)
    func_str, args_str, kwargs_str = enter_record.args
    if func_pattern:
        assert re.search(func_pattern, func_str)
    if args_pattern:
        assert re.search(args_pattern, args_str)
    if kwargs_pattern:
        assert re.search(kwargs_pattern, kwargs_str)

    leave_record = log_capture.records[1]
    assert leave_record.name == _EXP_LOGGER_NAME
    assert leave_record.levelname == _EXP_LOG_LEVEL
    assert re.match(_EXP_LOG_MSG_LEAVE_PATTERN, leave_record.msg)
    func_str, return_str = leave_record.args
    if func_pattern:
        assert re.search(func_pattern, func_str)
    if return_pattern:
        assert re.search(return_pattern, return_str)


def test_1a_global_from_global(capture):
    # pylint: disable=redefined-outer-name
    """Simple test calling a decorated global function from a global
    function."""

    call_from_global(decorated_global_function)

    assert_log_capture(
        capture, func_pattern='decorated_global_function')


def test_1b_global_from_method(capture):
    # pylint: disable=redefined-outer-name
    """Simple test calling a decorated global function from a method."""

    CallerClass().call_from_method(decorated_global_function)

    assert_log_capture(
        capture, func_pattern='decorated_global_function')


def test_1c_global_props_args_from_global(capture):
    # pylint: disable=redefined-outer-name
    """Simple test calling a decorated global function with properties as args,
    from a global function."""
    props = {
        'prop1': 'value1',
        'hideme': 'secret',
    }
    blanked_props = {
        'prop1': 'value1',
        'hideme': BLANKED_OUT_STRING,
    }
    call_from_global(decorated_global_props_function, props)

    assert_log_capture(
        capture, func_pattern='decorated_global_props_function',
        args_pattern=re.escape(str(blanked_props)),
        return_pattern=re.escape(str(props)))


def test_1c_global_props_kwargs_from_global(capture):
    # pylint: disable=redefined-outer-name
    """Simple test calling a decorated global function with properties as
    kwargs, from a global function."""

    call_from_global(decorated_global_props_function,
                     properties={'prop1': 'value1'})

    assert_log_capture(
        capture, func_pattern='decorated_global_props_function')


def test_2a_global_inner1_from_global(capture):
    # pylint: disable=redefined-outer-name
    """Simple test calling a decorated inner function defined in a global
    function from a global function."""

    decorated_inner1_function = get_decorated_inner1_function()

    call_from_global(decorated_inner1_function)

    assert_log_capture(
        capture, func_pattern='global1_function.decorated_inner1_function')


def test_2b_global_inner1_from_method(capture):
    # pylint: disable=redefined-outer-name
    """Simple test calling a decorated inner function defined in a global
    function from a method."""

    decorated_inner1_function = get_decorated_inner1_function()

    CallerClass().call_from_method(decorated_inner1_function)

    assert_log_capture(
        capture, func_pattern='global1_function.decorated_inner1_function')


def test_3a_global_inner2_from_global(capture):
    # pylint: disable=redefined-outer-name
    """Simple test calling a decorated inner function defined in an inner
    function defined in a global function from a global function."""

    decorated_inner2_function = get_decorated_inner2_function()

    call_from_global(decorated_inner2_function)

    assert_log_capture(
        capture, func_pattern='inner1_function.decorated_inner2_function')


def test_3b_global_inner1_from_method(capture):
    # pylint: disable=redefined-outer-name
    """Simple test calling a decorated inner function defined in an inner
    function defined in a global function from a method."""

    decorated_inner2_function = get_decorated_inner2_function()

    CallerClass().call_from_method(decorated_inner2_function)

    assert_log_capture(
        capture, func_pattern='inner1_function.decorated_inner2_function')


def test_4a_method_from_global(capture):
    # pylint: disable=redefined-outer-name
    """Simple test calling a decorated method from a global function."""

    decorated_method = Decorator1Class.decorated_method
    d = Decorator1Class()

    call_from_global(decorated_method, d)

    assert_log_capture(
        capture, func_pattern='Decorator1Class.decorated_method')


def test_4b_method_from_method(capture):
    # pylint: disable=redefined-outer-name
    """Simple test calling a decorated method from a method."""

    decorated_method = Decorator1Class.decorated_method
    d = Decorator1Class()

    CallerClass().call_from_method(decorated_method, d)

    assert_log_capture(
        capture, func_pattern='Decorator1Class.decorated_method')


def test_5a_method_from_global(capture):
    # pylint: disable=redefined-outer-name
    """Simple test calling a decorated inner function defined in a method
    from a global function."""

    decorated_inner_function = \
        Decorator2Class.get_decorated_inner_function()

    call_from_global(decorated_inner_function)

    assert_log_capture(
        capture, func_pattern='method.decorated_inner_function')


def test_5b_method_from_method(capture):
    # pylint: disable=redefined-outer-name
    """Simple test calling a decorated inner function defined in a method
    from a method."""

    decorated_inner_function = \
        Decorator2Class.get_decorated_inner_function()

    CallerClass().call_from_method(decorated_inner_function)

    assert_log_capture(
        capture, func_pattern='method.decorated_inner_function')


def test_decorated_class():
    # pylint: disable=unused-variable
    """Test that using the decorator on a class raises TypeError."""

    with pytest.raises(TypeError):

        @logged_api_call
        class DecoratedClass:
            # pylint: disable=too-few-public-methods
            """A decorated class"""
            pass


def test_decorated_property():
    # pylint: disable=unused-variable
    """Test that using the decorator on a property raises TypeError."""

    with pytest.raises(TypeError):

        class Class:
            # pylint: disable=too-few-public-methods
            """A class with a decorated property"""

            @logged_api_call
            @property
            def decorated_property(self):
                """A decorated property"""
                return self


def test_root_logger():
    """Test that get_logger('') returns the Python root logger."""

    py_logger = logging.getLogger()

    zhmc_logger = get_logger('')

    assert zhmc_logger is py_logger


def test_foo_logger():
    """Test that get_logger('zhmcclient.foo') returns the same-named
    Python logger and has at least one handler."""

    py_logger = logging.getLogger('zhmcclient.foo')

    zhmc_logger = get_logger('zhmcclient.foo')

    assert zhmc_logger is py_logger
    assert len(zhmc_logger.handlers) >= 1



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The :class:`~zhmcclient_mock.IdPool` class provides a pool of integer ID values
from a defined value range. This is used for example to manage automatically
allocated device numbers.
"""


__all__ = ['IdPool']


class IdPool:
    """
    A pool of integer ID values from a defined value range.

    The IDs can be allocated from and returned to the pool.

    The pool is optimized for memory consumption, by only materializing ID
    values as needed.
    """

    def __init__(self, lowest, highest):
        """
        Parameters:

          lowest (integer): Lowest value of the ID value range.

          highest (integer): Highest value of the ID value range.
        """

        if lowest > highest:
            raise ValueError(
                f"Lowest value {lowest} is higher than highest {highest}")

        # ID value range, using slice semantics (end points past the highest)
        self._range_start = lowest
        self._range_end = highest + 1

        # The ID values in use.
        self._used = set()

        # Free pool: The ID values that are free and materialized.
        self._free = set()

        # Start of new free ID values to be materialized when the free pool is
        # expanded.
        self._expand_start = lowest

        # Expansion chunk size: Number of new free ID values to be materialized
        # when the free pool is expanded.
        self._expand_len = 10

    def _expand(self):
        """
        Expand the free pool, if possible.

        If out of capacity w.r.t. the defined ID value range, ValueError is
        raised.
        """
        assert not self._free  # free pool is empty
        expand_end = self._expand_start + self._expand_len

        # Limit the expansion. This happens if the size of the value range is
        # not a multiple of the expansion chunk size.
        expand_end = min(expand_end, self._range_end)

        if self._expand_start == expand_end:
            raise ValueError("Out of capacity in ID pool")
        self._free = set(range(self._expand_start, expand_end))
        self._expand_start = expand_end

    def alloc(self):
        """
        Allocate an ID value and return it.

        Raises:

            ValueError: Out of capacity in ID pool.
        """
        if not self._free:
            self._expand()
        _id = self._free.pop()
        self._used.add(_id)
        return _id

    def free(self, id):
        # pylint: disable=redefined-builtin
        """
        Free an ID value.

        The ID value must be allocated.

        Raises:

            ValueError: ID value to be freed is not currently allocated.
        """
        self._free_impl(id, fail_if_not_allocated=True)

    def free_if_allocated(self, id):
        # pylint: disable=redefined-builtin
        """
        Free an ID value, if it is currently allocated.

        If the specified ID value is not currently allocated, nothing happens.
        """
        self._free_impl(id, fail_if_not_allocated=False)

    def _free_impl(self, id, fail_if_not_allocated):
        # pylint: disable=redefined-builtin
        """
        Implementation of free.
        """
        if id in self._used:
            self._used.remove(id)
            self._free.add(id)
        elif fail_if_not_allocated:
            raise ValueError(
                f"ID value to be freed is not currently allocated: {id}")



# Copyright 2023 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _group module.
"""


import re
import pytest

from zhmcclient import Client, HTTPError, NotFound, Group
from zhmcclient_mock import FakedSession
from tests.common.utils import assert_resources


class TestGroup:
    """All tests for the Group and GroupManager classes."""

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.

        Set up a faked session, and add a faked Console without any
        child resources.
        """
        # pylint: disable=attribute-defined-outside-init

        self.session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.client = Client(self.session)

        self.faked_console = self.session.hmc.consoles.add({
            'object-id': None,
            # object-uri will be automatically set
            'parent': None,
            'class': 'console',
            'name': 'fake-console1',
            'description': 'Console #1',
        })
        self.console = self.client.consoles.find(name=self.faked_console.name)

    def add_group(self, name):
        """
        Add a faked group object to the faked Console and return it.
        """
        faked_group = self.faked_console.groups.add({
            'object-id': f'oid-{name}',
            # object-uri will be automatically set
            'parent': None,
            'class': 'group',
            'name': name,
            'description': f'Group {name}',
        })
        return faked_group

    def test_group_manager_repr(self):
        """Test GroupManager.__repr__()."""

        group_mgr = self.console.groups

        # Execute the code to be tested
        repr_str = repr(group_mgr)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf'^{group_mgr.__class__.__name__}\s+at\s+'
            rf'0x{id(group_mgr):08x}\s+\(\\n.*',
            repr_str)

    def test_group_manager_initial_attrs(self):
        """Test initial attributes of GroupManager."""

        group_mgr = self.console.groups

        # Verify all public properties of the manager object
        assert group_mgr.resource_class == Group
        assert group_mgr.class_name == 'group'
        assert group_mgr.session is self.session
        assert group_mgr.parent is self.console
        assert group_mgr.console is self.console

    @pytest.mark.parametrize(
        "full_properties_kwargs, prop_names", [
            (dict(full_properties=False),
             ['object-uri', 'name']),
            (dict(full_properties=True),
             ['object-uri', 'name', 'description', 'match-info']),
            ({},  # test default for full_properties (False)
             ['object-uri', 'name']),
        ]
    )
    @pytest.mark.parametrize(
        "filter_args, exp_names", [
            (None,
             ['a', 'b']),
            ({},
             ['a', 'b']),
            ({'name': 'a'},
             ['a']),
        ]
    )
    def test_group_manager_list(
            self, filter_args, exp_names, full_properties_kwargs, prop_names):
        """Test GroupManager.list()."""

        faked_group1 = self.add_group(name='a')
        faked_group2 = self.add_group(name='b')
        faked_groups = [faked_group1, faked_group2]
        exp_faked_groups = [g for g in faked_groups if g.name in exp_names]
        group_mgr = self.console.groups

        # Execute the code to be tested
        groups = group_mgr.list(filter_args=filter_args,
                                **full_properties_kwargs)

        assert_resources(groups, exp_faked_groups, prop_names)

    @pytest.mark.parametrize(
        "input_props, exp_prop_names, exp_exc", [
            ({},  # props missing
             None,
             HTTPError({'http-status': 400, 'reason': 5})),
            ({'description': 'fake description X'},  # props missing
             None,
             HTTPError({'http-status': 400, 'reason': 5})),
            ({'description': 'fake description X',
              'name': 'a'},
             ['object-uri', 'name', 'description'],
             None),
        ]
    )
    def test_group_manager_create(self, input_props, exp_prop_names, exp_exc):
        """Test GroupManager.create()."""

        group_mgr = self.console.groups

        if exp_exc is not None:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                group_mgr.create(properties=input_props)

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

        else:

            # Execute the code to be tested.
            group = group_mgr.create(properties=input_props)

            # Check the resource for consistency within itself
            assert isinstance(group, Group)
            group_name = group.name
            exp_group_name = group.properties['name']
            assert group_name == exp_group_name
            group_uri = group.uri
            exp_group_uri = group.properties['object-uri']
            assert group_uri == exp_group_uri

            # Check the properties against the expected names and values
            for prop_name in exp_prop_names:
                assert prop_name in group.properties
                if prop_name in input_props:
                    value = group.properties[prop_name]
                    exp_value = input_props[prop_name]
                    assert value == exp_value

    def test_group_repr(self):
        """Test Group.__repr__()."""

        faked_group1 = self.add_group(name='a')
        group_mgr = self.console.groups
        group1 = group_mgr.find(name=faked_group1.name)

        # Execute the code to be tested
        repr_str = repr(group1)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf'^{group1.__class__.__name__}\s+at\s+'
            rf'0x{id(group1):08x}\s+\(\\n.*',
            repr_str)

    @pytest.mark.parametrize(
        "input_props, exp_exc", [
            ({'name': 'a'},
             None),
            ({'name': 'b'},
             None),
        ]
    )
    def test_group_delete(self, input_props, exp_exc):
        """Test Group.delete()."""

        faked_group = self.add_group(name=input_props['name'])

        group_mgr = self.console.groups
        group = group_mgr.find(name=faked_group.name)

        if exp_exc is not None:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                group.delete()

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

            # Check that the Group still exists
            group_mgr.find(name=faked_group.name)

        else:

            # Execute the code to be tested.
            group.delete()

            # Check that the Group no longer exists
            with pytest.raises(NotFound) as exc_info:
                group_mgr.find(name=faked_group.name)

    def test_group_delete_create_same_name(self):
        """Test Group.delete() followed by create() with same name."""

        group_name = 'faked_a'

        # Add the Group to be tested
        self.add_group(name=group_name)

        # Input properties for a Group with the same name
        sn_group_props = {
            'name': group_name,
            'description': 'Group with same name',
        }

        group_mgr = self.console.groups
        group = group_mgr.find(name=group_name)

        # Execute the deletion code to be tested
        group.delete()

        # Check that the Group no longer exists
        with pytest.raises(NotFound):
            group_mgr.find(name=group_name)

        # Execute the creation code to be tested.
        group_mgr.create(sn_group_props)

        # Check that the Group exists again under that name
        sn_group = group_mgr.find(name=group_name)
        description = sn_group.get_property('description')
        assert description == sn_group_props['description']



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=protected-access

"""
Unit tests for _resource module.
"""


import time
import re
import random
import threading
from collections import OrderedDict
from immutable_views import DictView
import pytest

from zhmcclient import BaseResource, BaseManager, Session, Client, \
    CeasedExistence
from zhmcclient._utils import divide_filter_args
from zhmcclient_mock import FakedSession


class MyResource(BaseResource):
    """
    A derived resource for testing the (abstract) BaseResource class.
    """

    # This init method is not part of the external API, so this testcase may
    # need to be updated if the API changes.
    def __init__(self, manager, uri, name, properties):
        # pylint: disable=useless-super-delegation
        super().__init__(manager, uri, name, properties)


class MyManager(BaseManager):
    """
    A derived resource manager for testing purposes.

    It is only needed because BaseResource needs it; it is not subject
    of test in this unit test module.
    """

    # This init method is not part of the external API, so this testcase may
    # need to be updated if the API changes.
    def __init__(self, session):
        super().__init__(
            resource_class=MyResource,
            class_name='myresource',
            session=session,
            parent=None,  # a top-level resource
            base_uri='/api/myresources',
            oid_prop='fake-oid-prop',
            uri_prop='fake-uri-prop',
            name_prop='fake-name-prop',
            query_props=['qp1', 'qp2'])

    def list(self, full_properties=False, filter_args=None):
        # We have this method here just to avoid the warning about
        # an unimplemented abstract method. It is not being used in this
        # set of testcases.
        raise NotImplementedError


class ResourceTestCase:
    """
    Base class for all tests in this file.
    """

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.
        """
        # pylint: disable=attribute-defined-outside-init

        self.session = Session(host='fake-host')
        self.mgr = MyManager(self.session)
        self.uri = self.mgr._base_uri + '/deadbeef-beef-beef-beef-deadbeefbeef'
        self.name = "fake-name"
        self.uri_prop = 'fake-uri-prop'  # same as in MyManager
        self.name_prop = 'fake-name-prop'  # same as in MyManager

    @staticmethod
    def assert_properties(resource, exp_props):
        """
        Assert that the properties of a resource object are as expected.
        """

        # Check the properties member type
        assert isinstance(resource.properties, DictView)

        # Verify that the resource properties are as expected
        assert len(resource.properties) == len(exp_props), (
            "Set of properties does not match. "
            f"Expected {resource.properties.keys()!r}, "
            f"got {exp_props.keys()!r}")

        for name, exp_value in exp_props.items():
            act_value = resource.properties[name]
            assert act_value == exp_value, (
                f"Property {name!r} does not match. "
                f"Expected {exp_value!r}, got {act_value!r}")


class TestInit(ResourceTestCase):
    """Test BaseResource initialization."""

    def test_empty_name(self):
        """Test with an empty set of input properties, with 'name'."""
        init_props = {}
        res_props = {
            self.uri_prop: self.uri,
            self.name_prop: self.name,
        }

        res = MyResource(self.mgr, self.uri, self.name, init_props)

        assert res.manager is self.mgr
        assert res.uri == self.uri
        assert res.name == self.name
        self.assert_properties(res, res_props)
        assert int(time.time()) - res.properties_timestamp <= 1
        assert res.full_properties is False

    def test_empty_no_name(self):
        """Test with an empty set of input properties, without 'name'."""
        init_props = {}
        res_props = {
            self.uri_prop: self.uri,
        }

        res = MyResource(self.mgr, self.uri, None, init_props)

        assert res.manager is self.mgr
        assert res.uri == self.uri
        self.assert_properties(res, res_props)
        assert int(time.time()) - res.properties_timestamp <= 1
        assert res.full_properties is False

    def test_simple(self):
        """Test with a simple set of input properties."""
        init_props = {
            'prop1': 'abc',
            'prop2': 100042
        }
        res_props = {
            self.uri_prop: self.uri,
            'prop1': 'abc',
            'prop2': 100042
        }

        res = MyResource(self.mgr, self.uri, None, init_props)

        assert res.manager is self.mgr
        assert res.uri == self.uri
        self.assert_properties(res, res_props)
        assert int(time.time()) - res.properties_timestamp <= 1
        assert res.full_properties is False

    def test_prop_case(self):
        """Test case sensitivity for the input properties."""
        init_props = {
            'prop1': 'abc',
            'Prop1': 100042,
        }
        res_props = {
            self.uri_prop: self.uri,
            'prop1': 'abc',
            'Prop1': 100042,
        }

        res = MyResource(self.mgr, self.uri, None, init_props)

        assert res.manager is self.mgr
        assert res.uri == self.uri
        self.assert_properties(res, res_props)
        assert int(time.time()) - res.properties_timestamp <= 1
        assert res.full_properties is False

    def test_invalid_type(self):
        """Test that input properties with an invalid type fail."""
        init_props = 42
        try:

            MyResource(self.mgr, self.uri, None, init_props)

        except TypeError:
            pass
        else:
            raise AssertionError(
                "TypeError was not raised when initializing resource "
                f"with invalid properties: {init_props!r}")

    def test_str(self):
        """Test BaseResource.__str__()."""
        init_props = {
            'prop1': 'abc',
            'Prop1': 100042,
        }
        resource = MyResource(self.mgr, self.uri, None, init_props)

        str_str = str(resource)

        str_str = str_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(rf'^{resource.__class__.__name__}\s*\(.*',
                        str_str)

    def test_repr(self):
        """Test BaseResource.__repr__()."""
        init_props = {
            'prop1': 'abc',
            'Prop1': 100042,
        }
        resource = MyResource(self.mgr, self.uri, None, init_props)

        repr_str = repr(resource)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf'^{resource.__class__.__name__}\s+at\s+'
            rf'0x{id(resource):08x}\s+\(\\n.*',
            repr_str)


class TestPropertySet(ResourceTestCase):
    """Test BaseResource by setting properties."""

    def test_add_to_empty(self):
        """Test setting a property in a resource object with no properties."""
        init_props = {}
        set_props = {
            'prop1': 'abc',
            'prop2': 100042,
        }
        res_props = {
            self.uri_prop: self.uri,
        }

        res = MyResource(self.mgr, self.uri, None, init_props)

        for key, value in set_props.items():
            # Since zhmcclient 0.31.0, the 'properties' attribute has type
            # DictView which prevents modifications to the dictionary.
            with pytest.raises(TypeError):
                res.properties[key] = value

        self.assert_properties(res, res_props)

    def test_replace_one_add_one(self):
        """Test replacing and adding a property in a resource object."""
        init_props = {
            'prop1': 42,
        }
        set_props = {
            'prop1': 'abc',
            'prop2': 100042,
        }
        res_props = {
            self.uri_prop: self.uri,
            'prop1': 42,
        }

        res = MyResource(self.mgr, self.uri, None, init_props)

        for key, value in set_props.items():
            # Since zhmcclient 0.31.0, the 'properties' attribute has type
            # DictView which prevents modifications to the dictionary.
            with pytest.raises(TypeError):
                res.properties[key] = value

        self.assert_properties(res, res_props)


class TestPropertyDel(ResourceTestCase):
    """Test BaseResource by deleting properties."""

    def test_del_one(self):
        """Test deleting a property in a resource object."""
        init_props = {
            'prop1': 'abc',
            'prop2': 100042,
        }
        del_keys = ('prop1',)
        res_props = {
            self.uri_prop: self.uri,
            'prop1': 'abc',
            'prop2': 100042,
        }

        res = MyResource(self.mgr, self.uri, None, init_props)

        for key in del_keys:
            # Since zhmcclient 0.31.0, the 'properties' attribute has type
            # DictView which prevents modifications to the dictionary.
            with pytest.raises(TypeError):
                del res.properties[key]

        self.assert_properties(res, res_props)

    def test_del_all_input(self):
        """Test deleting all input properties in a resource object."""
        init_props = {
            'prop1': 'abc',
            'prop2': 100042,
        }
        del_keys = ('prop1', 'prop2')
        res_props = {
            self.uri_prop: self.uri,
            'prop1': 'abc',
            'prop2': 100042,
        }

        res = MyResource(self.mgr, self.uri, None, init_props)

        for key in del_keys:
            # Since zhmcclient 0.31.0, the 'properties' attribute has type
            # DictView which prevents modifications to the dictionary.
            with pytest.raises(TypeError):
                del res.properties[key]

        self.assert_properties(res, res_props)

    def test_del_invalid(self):
        """Test deleting an invalid property in a resource object."""
        init_props = {
            'prop1': 'abc',
            'prop2': 100042,
        }
        res_props = {
            self.uri_prop: self.uri,
            'prop1': 'abc',
            'prop2': 100042,
        }

        res = MyResource(self.mgr, self.uri, None, init_props)

        invalid_key = 'inv1'
        # Rejection of deletion is checked byfore invalid key is checked.
        with pytest.raises(TypeError):
            del res.properties[invalid_key]

        self.assert_properties(res, res_props)

    def test_clear(self):
        """Test clearing the properties in a resource object."""
        init_props = {
            'prop1': 'abc',
            'prop2': 100042,
        }
        res_props = {
            self.uri_prop: self.uri,
            'prop1': 'abc',
            'prop2': 100042,
        }

        res = MyResource(self.mgr, self.uri, None, init_props)

        # Since zhmcclient 0.31.0, the 'properties' attribute has type
        # DictView which prevents modifications to the dictionary.
        with pytest.raises(AttributeError):
            res.properties.clear()

        self.assert_properties(res, res_props)


class TestManagerDivideFilter(ResourceTestCase):
    """
    Test the divide_filter_args() utils method (previously in BaseManager).
    """

    # Reserved chars are defined in RFC 3986 as gen-delims and sub-delims.
    reserved_chars = [
        ':', '/', '?', '#', '[', ']', '@',  # gen-delims
        '!', '$', '&', "'", '(', ')', '*', '+', ',', ';', '='  # sub-delims
    ]

    # Percent-escapes for the reserved chars, in the same order.
    reserved_escapes = [
        '%3A', '%2F', '%3F', '%23', '%5B', '%5D', '%40',  # gen-delims
        '%21', '%24', '%26', '%27', '%28', '%29', '%2A',  # sub-delims
        '%2B', '%2C', '%3B', '%3D',  # sub-delims
    ]

    def test_none(self):
        """Test with None as filter arguments."""
        filter_args = None

        parm_str, cf_args = divide_filter_args(
            self.mgr._query_props, filter_args)

        assert parm_str == []
        assert cf_args == {}

    def test_empty(self):
        """Test with an empty set of filter arguments."""
        filter_args = {}

        parm_str, cf_args = divide_filter_args(
            self.mgr._query_props, filter_args)

        assert parm_str == []
        assert cf_args == {}

    def test_one_string_qp(self):
        """Test with one string filter argument that is a query parm."""
        filter_args = {'qp1': 'bar'}

        parm_str, cf_args = divide_filter_args(
            self.mgr._query_props, filter_args)

        assert parm_str == ['qp1=bar']
        assert cf_args == {}

    def test_one_string_cf(self):
        """Test with one string filter argument that is a client filter."""
        filter_args = {'foo': 'bar'}

        parm_str, cf_args = divide_filter_args(
            self.mgr._query_props, filter_args)

        assert parm_str == []
        assert cf_args == {'foo': 'bar'}

    def test_one_integer_qp(self):
        """Test with one integer filter argument that is a query parm."""
        filter_args = {'qp2': 42}

        parm_str, cf_args = divide_filter_args(
            self.mgr._query_props, filter_args)

        assert parm_str == ['qp2=42']
        assert cf_args == {}

    def test_one_integer_cf(self):
        """Test with one integer filter argument that is a client filter."""
        filter_args = {'foo': 42}

        parm_str, cf_args = divide_filter_args(
            self.mgr._query_props, filter_args)

        assert parm_str == []
        assert cf_args == {'foo': 42}

    def test_one_str_reserved_val_qp(self):
        """Test with one string filter argument with reserved URI chars in
        its value that is a query parm."""
        char_str = '_'.join(self.reserved_chars)
        escape_str = '_'.join(self.reserved_escapes)
        filter_args = {'qp1': char_str}

        parm_str, cf_args = divide_filter_args(
            self.mgr._query_props, filter_args)

        assert parm_str == [f'qp1={escape_str}']
        assert cf_args == {}

    def test_one_str_reserved_val_cf(self):
        """Test with one string filter argument with reserved URI chars in
        its value that is a client filter."""
        char_str = '_'.join(self.reserved_chars)
        filter_args = {'foo': char_str}

        parm_str, cf_args = divide_filter_args(
            self.mgr._query_props, filter_args)

        assert parm_str == []
        assert cf_args == {'foo': char_str}

    def test_one_str_dash_name_qp(self):
        """Test with one string filter argument with a dash in its name that is
        a query parm."""
        filter_args = {'foo-boo': 'bar'}
        self.mgr._query_props.append('foo-boo')

        parm_str, cf_args = divide_filter_args(
            self.mgr._query_props, filter_args)

        assert parm_str == ['foo-boo=bar']
        assert cf_args == {}

    def test_one_str_reserved_name_qp(self):
        """Test with one string filter argument with reserved URI chars in
        its name that is a query parm."""
        char_str = '_'.join(self.reserved_chars)
        escape_str = '_'.join(self.reserved_escapes)
        filter_args = {char_str: 'bar'}
        self.mgr._query_props.append(char_str)

        parm_str, cf_args = divide_filter_args(
            self.mgr._query_props, filter_args)

        assert parm_str == [f'{escape_str}=bar']
        assert cf_args == {}

    def test_two_qp(self):
        """Test with two filter arguments that are query parms."""
        filter_args = OrderedDict([('qp1', 'bar'), ('qp2', 42)])

        parm_str, cf_args = divide_filter_args(
            self.mgr._query_props, filter_args)

        assert parm_str == ['qp1=bar', 'qp2=42']
        assert cf_args == {}

    def test_two_qp_cf(self):
        """Test with two filter arguments where one is a query parm and one is
        a client filter."""
        filter_args = OrderedDict([('qp1', 'bar'), ('foo', 42)])

        parm_str, cf_args = divide_filter_args(
            self.mgr._query_props, filter_args)

        assert parm_str == ['qp1=bar']
        assert cf_args == {'foo': 42}

    def test_two_cf_qp(self):
        """Test with two filter arguments where one is a client filter and one
        is a query parm."""
        filter_args = OrderedDict([('foo', 'bar'), ('qp1', 42)])

        parm_str, cf_args = divide_filter_args(
            self.mgr._query_props, filter_args)

        assert parm_str == ['qp1=42']
        assert cf_args == {'foo': 'bar'}

    def test_two_two_qp(self):
        """Test with two filter arguments, one of which is a list of two, and
        both are query parms."""
        filter_args = OrderedDict([('qp1', 'bar'), ('qp2', [42, 7])])

        parm_str, cf_args = divide_filter_args(
            self.mgr._query_props, filter_args)

        assert parm_str == ['qp1=bar', 'qp2=42', 'qp2=7']
        assert cf_args == {}

    def test_two_str_reserved_val_qp(self):
        """Test with two filter arguments, one of which is a list of two, and
        has reserved URI chars, and both are query parms."""
        char_str = '_'.join(self.reserved_chars)
        escape_str = '_'.join(self.reserved_escapes)
        filter_args = OrderedDict([('qp1', 'bar'), ('qp2', [42, char_str])])

        parm_str, cf_args = divide_filter_args(
            self.mgr._query_props, filter_args)

        assert parm_str == ['qp1=bar', 'qp2=42', f'qp2={escape_str}']
        assert cf_args == {}


class TestThreadingSerialization(ResourceTestCase):
    """
    Test serialization of resource property update/access.
    """

    def test_serialization_1(self):
        """
        Test serialization of resource property update/access.
        """

        resource = MyResource(self.mgr, self.uri, 'res1', dict(p1=1, p2=1))

        def update():
            for _ in range(0, 10000):
                value = random.randint(1, 100)
                props = dict(p1=value, p2=value)
                resource.update_properties_local(props)

        def get_assert():
            for _ in range(0, 10000):
                value1, value2 = resource.get_properties_local(['p1', 'p2'])
                assert value1 == value2

        threads = []
        for _ in range(0, 10):

            thread = threading.Thread(target=update)
            thread.start()
            threads.append(thread)

            thread = threading.Thread(target=get_assert)
            thread.start()
            threads.append(thread)

        for thread in threads:
            thread.join()

    TESTCASES_GET_PROPERTIES_LOCAL = [
        # Testcases for test_get_properties_local().
        # Each list item is a tuple defining a testcase in the following format:
        # - desc: Testcase description
        # - create_properties: Initial properties for test resource
        # - get_properties: properties parm for tested function
        # - default: default parm for tested function
        # - exp_values: Expected properties returned by tested function
        (
            "Get two properties that exist",
            {
                'p1': 'v1',
                'p2': 'v2',
            },
            ['p1', 'p2'],
            'vd',
            ['v1', 'v2'],
        ),
        (
            "Get two properties that do not exist, using single default value",
            {},
            ['p1', 'p2'],
            'vd',
            ['vd', 'vd'],
        ),
        (
            "Get two properties that do not exist, using two default values",
            {},
            ['p1', 'p2'],
            ['vd1', 'vd2'],
            ['vd1', 'vd2'],
        ),
    ]

    @pytest.mark.parametrize(
        "desc, create_properties, get_properties, default, exp_values",
        TESTCASES_GET_PROPERTIES_LOCAL)
    def test_get_properties_local(
            self, desc, create_properties, get_properties, default, exp_values):
        # pylint: disable=unused-argument
        """
        Test get_properties_local().
        """

        resource = MyResource(self.mgr, self.uri, 'res1', create_properties)

        values = resource.get_properties_local(get_properties, default)

        for i, name in enumerate(get_properties):
            value = values[i]
            exp_value = exp_values[i]

            assert value == exp_value, \
                f"Unexpected property value for '{name}'"


class TestPropertyMethodsMocked:
    """
    All tests of property methods that need mocked resources.
    """

    RESOURCE_OID = 'res-oid'
    RESOURCE_NAME = 'res-name'
    RESOURCE_DESC = 'Resource Description'

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.

        Set up a faked session, and add a faked CPC in DPM mode without any
        child resources.
        """
        # pylint: disable=attribute-defined-outside-init

        self.session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.client = Client(self.session)

        self.faked_cpc = self.session.hmc.cpcs.add({
            'object-id': 'fake-cpc1-oid',
            # object-uri is set up automatically
            'parent': None,
            'class': 'cpc',
            'name': 'fake-cpc1-name',
            'description': 'CPC #1 (DPM mode)',
            'status': 'active',
            'dpm-enabled': True,
            'is-ensemble-member': False,
            'iml-mode': 'dpm',
            'machine-type': '2964',  # z13
            'machine-model': 'N10',
        })
        self.cpc = self.client.cpcs.find(name='fake-cpc1-name')

    def add_standard_hipersocket(self):
        """
        Add a standard Hipersocket adapter with one port to the faked HMC.

        Adapter resources do not support the 'properties' query parameter.
        """

        # Adapter properties that will be auto-set:
        # - object-uri
        # - adapter-family
        # - network-port-uris (to empty array)
        faked_hs = self.faked_cpc.adapters.add({
            'object-id': self.RESOURCE_OID,
            'parent': self.faked_cpc.uri,
            'class': 'adapter',
            'name': self.RESOURCE_NAME,
            'description': self.RESOURCE_DESC,
            'status': 'active',
            'type': 'hipersockets',
            'adapter-id': '123',
            'detected-card-type': 'hipersockets',
            'port-count': 1,
            'network-port-uris': [],
            'state': 'online',
            'configured-capacity': 32,
            'used-capacity': 0,
            'allowed-capacity': 32,
            'maximum-total-capacity': 32,
            'physical-channel-status': 'operating',
            'maximum-transmission-unit-size': 56,
        })

        # Port properties that will be auto-set:
        # - element-uri
        # Properties in parent adapter that will be auto-set:
        # - network-port-uris
        faked_hs.ports.add({
            'element-id': 'fake-port1-oid',
            'parent': faked_hs.uri,
            'class': 'network-port',
            'index': 0,
            'name': 'fake-port1-name',
            'description': 'Hipersocket #1 Port #1',
        })
        return faked_hs

    def add_standard_partition(self):
        """
        Add a standard partition.

        Partition resources support the 'properties' query parameter.
        """

        # Partition properties that will be auto-set:
        # - object-uri
        faked_part = self.faked_cpc.partitions.add({
            'object-id': self.RESOURCE_OID,
            'parent': self.faked_cpc.uri,
            'class': 'partition',
            'name': self.RESOURCE_NAME,
            'description': self.RESOURCE_DESC,
            'status': 'stopped',
        })
        return faked_part

    @pytest.mark.parametrize(
        # Indicates whether to delete the resource before test
        "delete", [False, True])
    def test_pull_full_properties(self, delete):
        """
        Test BaseResource.pull_full_properties().
        """
        faked_res = self.add_standard_hipersocket()
        res_mgr = self.cpc.adapters
        resource = res_mgr.find(name=faked_res.name)
        propnames_find = set(resource.properties.keys())
        assert resource.full_properties is False

        if delete:
            # Delete the resource on the HMC, without affecting the state of
            # the local zhmcclient resource object. This simulates some other
            # user who has deleted the resource on the HMC while we have it
            # here as a zhmcclient resource object.
            resource.manager.session.delete(resource.uri)

            with pytest.raises(CeasedExistence):

                # The code to be tested
                resource.pull_full_properties()

        else:

            # The code to be tested
            resource.pull_full_properties()

            assert resource.full_properties is True
            propnames_pull = set(resource.properties.keys())
            assert propnames_pull.issuperset(propnames_find)
            assert dict(resource.properties) == faked_res.properties

    TESTCASES_PULL_PROPERTIES = [
        # Testcases for test_pull_properties().
        # Each list item is a tuple defining a testcase in the following format:
        # - desc: Testcase description
        # - input_kwargs: keyword input parameters for the function
        # - delete: Indicates whether to delete the resource before test
        # - supports_properties: Resource supports 'properties' query parm
        # - exp_propnames: Expected property names in resource after test
        # - exp_full_props: Expected value for full_properties attribute
        # - exp_exc_type: Expected type of exception, or None for success

        (
            "No properties requested (None); "
            "resource exists and does not support 'properties' query parm",
            dict(properties=None),
            False,
            False,
            {'name', 'object-uri'},
            False,
            None,
        ),
        (
            "No properties requested (None); "
            "resource exists and supports 'properties' query parm",
            dict(properties=None),
            False,
            True,
            {'name', 'object-uri'},
            False,
            None,
        ),
        (
            "No properties requested (None); "
            "resource is deleted and does not support 'properties' query parm",
            dict(properties=None),
            True,
            False,
            {'name', 'object-uri'},
            False,
            None,
        ),
        (
            "No properties requested (None); "
            "resource is deleted and supports 'properties' query parm",
            dict(properties=None),
            True,
            True,
            {'name', 'object-uri'},
            False,
            None,
        ),

        (
            "No properties requested (empty list); "
            "resource exists and does not support 'properties' query parm",
            dict(properties=[]),
            False,
            False,
            {'name', 'object-uri'},
            False,
            None,
        ),
        (
            "No properties requested (empty list); "
            "resource exists and supports 'properties' query parm",
            dict(properties=[]),
            False,
            True,
            {'name', 'object-uri'},
            False,
            None,
        ),
        (
            "No properties requested (empty list); "
            "resource is deleted and does not support 'properties' query parm",
            dict(properties=[]),
            True,
            False,
            {'name', 'object-uri'},
            False,
            None,
        ),
        (
            "No properties requested (empty list); "
            "resource is deleted and supports 'properties' query parm",
            dict(properties=[]),
            True,
            True,
            {'name', 'object-uri'},
            False,
            None,
        ),

        (
            "One valid property requested; "
            "resource exists and does not support 'properties' query parm",
            dict(properties=['object-id']),
            False,
            False,
            {'name', 'object-uri', 'object-id'},
            True,
            None,
        ),
        (
            "One valid property requested; "
            "resource exists and supports 'properties' query parm",
            dict(properties=['object-id']),
            False,
            True,
            {'name', 'object-uri', 'object-id'},
            False,
            None,
        ),
        (
            "One valid property requested; "
            "resource is deleted and does not support 'properties' query parm",
            dict(properties=['object-id']),
            True,
            False,
            None,
            None,
            CeasedExistence,
        ),
        (
            "One valid property requested; "
            "resource is deleted and supports 'properties' query parm",
            dict(properties=['object-id']),
            True,
            True,
            None,
            None,
            CeasedExistence,
        ),

        (
            "One invalid property requested; "
            "resource exists and does not support 'properties' query parm",
            dict(properties=['invalid-property']),
            False,
            False,
            {'name', 'object-uri', 'object-id'},
            True,
            None,
        ),
        (
            "One invalid property requested; "
            "resource exists and supports 'properties' query parm",
            dict(properties=['invalid-property']),
            False,
            True,
            {'name', 'object-uri', 'object-id'},
            True,
            None,
        ),
        (
            "One invalid property requested; "
            "resource is deleted and does not support 'properties' query parm",
            dict(properties=['invalid-property']),
            True,
            False,
            None,
            None,
            CeasedExistence,
        ),
        (
            "One invalid property requested; "
            "resource is deleted and supports 'properties' query parm",
            dict(properties=['invalid-property']),
            True,
            True,
            None,
            None,
            CeasedExistence,
        ),
    ]

    @pytest.mark.parametrize(
        "desc, input_kwargs, delete, supports_properties, exp_propnames, "
        "exp_full_props, exp_exc_type",
        TESTCASES_PULL_PROPERTIES)
    def test_pull_properties(
            self, desc, input_kwargs, delete, supports_properties,
            exp_propnames, exp_full_props, exp_exc_type):
        # pylint: disable=unused-argument
        """
        Test BaseResource.pull_properties().
        """
        if supports_properties:
            # Partitions support the properties query parm
            faked_res = self.add_standard_partition()
            res_mgr = self.cpc.partitions
        else:
            # Adapters do not support the properties query parm
            faked_res = self.add_standard_hipersocket()
            res_mgr = self.cpc.adapters
        assert res_mgr.supports_properties == supports_properties

        resource = res_mgr.find(name=faked_res.name)
        assert resource.full_properties is False

        if delete:
            # Delete the resource on the HMC, without affecting the state of
            # the local zhmcclient resource object. This simulates some other
            # user who has deleted the resource on the HMC while we have it
            # here as a zhmcclient resource object.
            resource.manager.session.delete(resource.uri)

        if exp_exc_type:
            with pytest.raises(exp_exc_type) as exc_info:

                # The code to be tested
                resource.pull_properties(**input_kwargs)

            exc = exc_info.value
            assert isinstance(exc, exp_exc_type)
        else:

            # The code to be tested
            resource.pull_properties(**input_kwargs)

            propnames = set(resource.properties.keys())
            assert propnames.issuperset(exp_propnames)

            assert resource.full_properties == exp_full_props

    TESTCASES_GET_PROPERTY = [
        # Testcases for test_get_property().
        # Each list item is a tuple defining a testcase in the following format:
        # - desc: Testcase description
        # - input_kwargs: keyword input parameters for the function
        # - delete: Indicates whether to delete the resource before test
        # - supports_properties: Resource supports 'properties' query parm
        # - exp_value: Expected return value of the function
        # - exp_propnames: Expected property names in resource after test
        # - exp_full_props: Expected value for full_properties attribute
        # - exp_exc_type: Expected type of exception, or None for success

        (
            "Valid locally existing property requested; "
            "resource exists and does not support 'properties' query parm",
            dict(name='name'),
            False,
            False,
            RESOURCE_NAME,
            {'name', 'object-uri'},
            False,
            None,
        ),
        (
            "Valid locally existing property requested; "
            "resource exists and supports 'properties' query parm",
            dict(name='name'),
            False,
            True,
            RESOURCE_NAME,
            {'name', 'object-uri'},
            False,
            None,
        ),
        (
            "Valid locally existing property requested; "
            "resource is deleted and does not support 'properties' query parm",
            dict(name='name'),
            True,
            False,
            RESOURCE_NAME,
            {'name', 'object-uri'},
            False,
            None,
        ),
        (
            "Valid locally existing property requested; "
            "resource is deleted and supports 'properties' query parm",
            dict(name='name'),
            True,
            True,
            RESOURCE_NAME,
            {'name', 'object-uri'},
            False,
            None,
        ),

        (
            "Valid not locally existing property requested; "
            "resource exists and does not support 'properties' query parm",
            dict(name='object-id'),
            False,
            False,
            RESOURCE_OID,
            {'name', 'object-uri', 'object-id'},
            False,
            None,
        ),
        (
            "Valid not locally existing property requested; "
            "resource exists and supports 'properties' query parm",
            dict(name='object-id'),
            False,
            True,
            RESOURCE_OID,
            {'name', 'object-uri', 'object-id'},
            False,
            None,
        ),
        (
            "Valid not locally existing property requested; "
            "resource is deleted and does not support 'properties' query parm",
            dict(name='object-id'),
            True,
            False,
            RESOURCE_OID,
            {'name', 'object-uri', 'object-id'},
            False,
            None,
        ),
        (
            "Valid not locally existing property requested; "
            "resource is deleted and supports 'properties' query parm",
            dict(name='object-id'),
            True,
            True,
            RESOURCE_OID,
            {'name', 'object-uri', 'object-id'},
            False,
            None,
        ),

        (
            "Invalid property requested; "
            "resource exists and does not support 'properties' query parm",
            dict(name='invalid-property'),
            False,
            False,
            None,
            None,
            None,
            KeyError,
        ),
        (
            "Invalid property requested; "
            "resource exists and supports 'properties' query parm",
            dict(name='invalid-property'),
            False,
            True,
            None,
            None,
            None,
            KeyError,
        ),
        (
            "Invalid property requested; "
            "resource is deleted and does not support 'properties' query parm",
            dict(name='invalid-property'),
            True,
            False,
            None,
            None,
            None,
            CeasedExistence,
        ),
        (
            "Invalid property requested; "
            "resource is deleted and supports 'properties' query parm",
            dict(name='invalid-property'),
            True,
            True,
            None,
            None,
            None,
            CeasedExistence,
        ),
    ]

    @pytest.mark.parametrize(
        "desc, input_kwargs, delete, supports_properties, exp_value, "
        "exp_propnames, exp_full_props, exp_exc_type",
        TESTCASES_GET_PROPERTY)
    def test_get_property(
            self, desc, input_kwargs, delete, supports_properties, exp_value,
            exp_propnames, exp_full_props, exp_exc_type):
        # pylint: disable=unused-argument
        """
        Test BaseResource.get_property().
        """
        if supports_properties:
            # Partitions support the properties query parm
            faked_res = self.add_standard_partition()
            res_mgr = self.cpc.partitions
        else:
            # Adapters do not support the properties query parm
            faked_res = self.add_standard_hipersocket()
            res_mgr = self.cpc.adapters
        assert res_mgr.supports_properties == supports_properties

        resource = res_mgr.find(name=faked_res.name)
        assert resource.full_properties is False

        if delete:
            # Delete the resource on the HMC, without affecting the state of
            # the local zhmcclient resource object. This simulates some other
            # user who has deleted the resource on the HMC while we have it
            # here as a zhmcclient resource object.
            resource.manager.session.delete(resource.uri)

        if exp_exc_type:
            with pytest.raises(exp_exc_type) as exc_info:

                # The code to be tested
                resource.get_property(**input_kwargs)

            exc = exc_info.value
            assert isinstance(exc, exp_exc_type)
        else:

            # The code to be tested
            value = resource.get_property(**input_kwargs)

            assert value == exp_value

            propnames = set(resource.properties.keys())
            assert propnames.issuperset(exp_propnames)

            assert resource.full_properties == exp_full_props

    TESTCASES_PROP = [
        # Testcases for test_prop().
        # Each list item is a tuple defining a testcase in the following format:
        # - desc: Testcase description
        # - input_kwargs: keyword input parameters for the function
        # - delete: Indicates whether to delete the resource before test
        # - supports_properties: Resource supports 'properties' query parm
        # - exp_value: Expected return value of the function
        # - exp_propnames: Expected property names in resource after test
        # - exp_full_props: Expected value for full_properties attribute
        # - exp_exc_type: Expected type of exception, or None for success

        (
            "Valid locally existing property requested without default; "
            "resource exists and does not support 'properties' query parm",
            dict(name='name'),
            False,
            False,
            RESOURCE_NAME,
            {'name', 'object-uri'},
            False,
            None,
        ),
        (
            "Valid locally existing property requested without default; "
            "resource exists and supports 'properties' query parm",
            dict(name='name'),
            False,
            True,
            RESOURCE_NAME,
            {'name', 'object-uri'},
            False,
            None,
        ),
        (
            "Valid locally existing property requested without default; "
            "resource is deleted and does not support 'properties' query parm",
            dict(name='name'),
            True,
            False,
            RESOURCE_NAME,
            {'name', 'object-uri'},
            False,
            None,
        ),
        (
            "Valid locally existing property requested without default; "
            "resource is deleted and supports 'properties' query parm",
            dict(name='name'),
            True,
            True,
            RESOURCE_NAME,
            {'name', 'object-uri'},
            False,
            None,
        ),

        (
            "Valid locally existing property requested with default; "
            "resource exists and does not support 'properties' query parm",
            dict(name='name', default='foo'),
            False,
            False,
            RESOURCE_NAME,
            {'name', 'object-uri'},
            False,
            None,
        ),
        (
            "Valid locally existing property requested with default; "
            "resource exists and supports 'properties' query parm",
            dict(name='name', default='foo'),
            False,
            True,
            RESOURCE_NAME,
            {'name', 'object-uri'},
            False,
            None,
        ),
        (
            "Valid locally existing property requested with default; "
            "resource is deleted and does not support 'properties' query parm",
            dict(name='name', default='foo'),
            True,
            False,
            RESOURCE_NAME,
            {'name', 'object-uri'},
            False,
            None,
        ),
        (
            "Valid locally existing property requested with default; "
            "resource is deleted and supports 'properties' query parm",
            dict(name='name', default='foo'),
            True,
            True,
            RESOURCE_NAME,
            {'name', 'object-uri'},
            False,
            None,
        ),

        (
            "Valid not locally existing property requested without default; "
            "resource exists and does not support 'properties' query parm",
            dict(name='object-id'),
            False,
            False,
            RESOURCE_OID,
            {'name', 'object-uri', 'object-id'},
            False,
            None,
        ),
        (
            "Valid not locally existing property requested without default; "
            "resource exists and supports 'properties' query parm",
            dict(name='object-id'),
            False,
            True,
            RESOURCE_OID,
            {'name', 'object-uri', 'object-id'},
            False,
            None,
        ),
        (
            "Valid not locally existing property requested without default; "
            "resource is deleted and does not support 'properties' query parm",
            dict(name='object-id'),
            True,
            False,
            RESOURCE_OID,
            {'name', 'object-uri', 'object-id'},
            False,
            None,
        ),
        (
            "Valid not locally existing property requested without default; "
            "resource is deleted and supports 'properties' query parm",
            dict(name='object-id'),
            True,
            True,
            RESOURCE_OID,
            {'name', 'object-uri', 'object-id'},
            False,
            None,
        ),

        (
            "Valid not locally existing property requested with default; "
            "resource exists and does not support 'properties' query parm",
            dict(name='object-id', default='foo'),
            False,
            False,
            RESOURCE_OID,
            {'name', 'object-uri', 'object-id'},
            False,
            None,
        ),
        (
            "Valid not locally existing property requested with default; "
            "resource exists and supports 'properties' query parm",
            dict(name='object-id', default='foo'),
            False,
            True,
            RESOURCE_OID,
            {'name', 'object-uri', 'object-id'},
            False,
            None,
        ),
        (
            "Valid not locally existing property requested with default; "
            "resource is deleted and does not support 'properties' query parm",
            dict(name='object-id', default='foo'),
            True,
            False,
            RESOURCE_OID,
            {'name', 'object-uri', 'object-id'},
            False,
            None,
        ),
        (
            "Valid not locally existing property requested with default; "
            "resource is deleted and supports 'properties' query parm",
            dict(name='object-id', default='foo'),
            True,
            True,
            RESOURCE_OID,
            {'name', 'object-uri', 'object-id'},
            False,
            None,
        ),

        (
            "Invalid property requested without default; "
            "resource exists and does not support 'properties' query parm",
            dict(name='invalid-property'),
            False,
            False,
            None,
            {'name', 'object-uri', 'invalid-property'},
            True,
            None,
        ),
        (
            "Invalid property requested without default; "
            "resource exists and supports 'properties' query parm",
            dict(name='invalid-property'),
            False,
            True,
            None,
            {'name', 'object-uri', 'invalid-property'},
            True,
            None,
        ),
        (
            "Invalid property requested without default; "
            "resource is deleted and does not support 'properties' query parm",
            dict(name='invalid-property'),
            True,
            False,
            None,
            None,
            None,
            CeasedExistence,
        ),
        (
            "Invalid property requested without default; "
            "resource is deleted and supports 'properties' query parm",
            dict(name='invalid-property'),
            True,
            True,
            None,
            None,
            None,
            CeasedExistence,
        ),

        (
            "Invalid property requested with default; "
            "resource exists and does not support 'properties' query parm",
            dict(name='invalid-property', default='foo'),
            False,
            False,
            'foo',
            {'name', 'object-uri', 'invalid-property'},
            True,
            None,
        ),
        (
            "Invalid property requested with default; "
            "resource exists and supports 'properties' query parm",
            dict(name='invalid-property', default='foo'),
            False,
            True,
            'foo',
            {'name', 'object-uri', 'invalid-property'},
            True,
            None,
        ),
        (
            "Invalid property requested with default; "
            "resource is deleted and does not support 'properties' query parm",
            dict(name='invalid-property', default='foo'),
            True,
            False,
            None,
            None,
            None,
            CeasedExistence,
        ),
        (
            "Invalid property requested with default; "
            "resource is deleted and supports 'properties' query parm",
            dict(name='invalid-property', default='foo'),
            True,
            True,
            None,
            None,
            None,
            CeasedExistence,
        ),
    ]

    @pytest.mark.parametrize(
        "desc, input_kwargs, delete, supports_properties, exp_value, "
        "exp_propnames, exp_full_props, exp_exc_type",
        TESTCASES_PROP)
    def test_prop(
            self, desc, input_kwargs, delete, supports_properties, exp_value,
            exp_propnames, exp_full_props, exp_exc_type):
        # pylint: disable=unused-argument
        """
        Test BaseResource.prop().
        """
        if supports_properties:
            # Partitions support the properties query parm
            faked_res = self.add_standard_partition()
            res_mgr = self.cpc.partitions
        else:
            # Adapters do not support the properties query parm
            faked_res = self.add_standard_hipersocket()
            res_mgr = self.cpc.adapters
        assert res_mgr.supports_properties == supports_properties

        resource = res_mgr.find(name=faked_res.name)
        assert resource.full_properties is False

        if delete:
            # Delete the resource on the HMC, without affecting the state of
            # the local zhmcclient resource object. This simulates some other
            # user who has deleted the resource on the HMC while we have it
            # here as a zhmcclient resource object.
            resource.manager.session.delete(resource.uri)

        if exp_exc_type:
            with pytest.raises(exp_exc_type) as exc_info:

                # The code to be tested
                resource.prop(**input_kwargs)

            exc = exc_info.value
            assert isinstance(exc, exp_exc_type)
        else:

            # The code to be tested
            value = resource.prop(**input_kwargs)

            assert value == exp_value

            propnames = set(resource.properties.keys())
            # prop() always adds the requested property
            if 'name' in input_kwargs:
                propnames.add(input_kwargs['name'])
            assert propnames.issuperset(exp_propnames)

            assert resource.full_properties == exp_full_props

    # TODO: Add test function for get_properties_pulled()
